import OpenGL.GL as gl
import OpenGL.arrays.vbo as vbo

import collections
import numpy
import math

"""Module to implement some basic 2D geometry classes and
routines. Written for E91 Humanoid Robotics, Spring 2013, by Matt
Zucker."""

######################################################################

def clamp(x, lo, hi):
    """Clamp a value to a 1D interval."""
    if lo > hi:
        return x
    else:
        return max(lo, min(x, hi))

def angleDiff(a, b):
    """Return the difference of a and b, wrapped into the [-pi,pi] interval."""
    return wrap(b-a, -math.pi, math.pi)

def wrap(value, lower, upper):
    """Wrap a value to a 1D interval. E.g. wrap(3.1, 0, 1.5) == 0.1"""
    if lower > upper:
        return value
    elif lower == upper:
        return lower
    d = upper - lower
    if value < lower:
        return upper - math.fmod(lower - value, d)
    elif value > upper:
        return lower + math.fmod(value - upper, d)
    else:
        return value

def point2D(*args):
    """Construct a numpy array via point2D(x,y) or point2D([x,y]) or
    point2D((x,y))."""
    if len(args) == 2:
        return numpy.array((args[0], args[1]), dtype='f')
    else:
        return numpy.array(args[0], dtype='f')

def dist2D(p1, p2):
    """Return the 2D distance between the two input vectors."""
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    return math.sqrt(dx*dx + dy*dy)

def distSqr2D(p1, p2):
    """Return the square of the 2D distance between the two input vectors."""
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    return dx*dx + dy*dy

def norm2D(p):
    """Return the 2D norm of the input vector."""
    return math.sqrt(p[0]*p[0] + p[1]*p[1])

def dot2D(a,b):
    """Return the 2D dot product of the two input vectors."""
    return a[0]*b[0] + a[1]*b[1]

def normSqr2D(p):
    """Return the square of the 2D norm of the input vector."""
    return p[0]*p[0] + p[1]*p[1]

def point2DFromAngle(theta):
    """Return an (x,y) pair corresponding to (cos(theta), sin(theta))."""
    return point2D(math.cos(theta), math.sin(theta))

def point2DFromPolar(r, theta):
    """Return a (x,y) pair corresponding to (r*cos(theta), r*sin(theta))."""
    return point2D(r*math.cos(theta), r*math.sin(theta))

def angle2D(p):
    return math.atan2(p[1], p[0])

def crossProduct2D(a, b):
    """Return the magnitude (z-component) of the cross product of two
    (x,y) pairs."""
    return a[0]*b[1] - a[1]*b[0]

def dir2D(p1, p2):
    """Return a normalized direction vector pointing from p1 to p2."""
    dx = p2[0]-p1[0]
    dy = p2[1]-p1[1]
    scl = 1.0/math.sqrt(dx*dx + dy*dy)
    return (dx*scl,dy*scl)

def normal2D(p1, p2):
    """Return the normalized vector pointing perpendicular to the
    vector from p1 to p2."""
    dx = p2[0]-p1[0]
    dy = p2[1]-p1[1]
    scl = 1.0/math.sqrt(dx*dx + dy*dy)
    return point2D(dy*scl,-dx*scl)

def perp2D(p):
    """Return a vector (y,-x) corresponding to the perpendicular to
    the vector input."""
    return point2D(p[1], -p[0])

def randpm(scl=1, *args):
    """Return a random number in the range of [-scl,+scl)."""
    return (numpy.random.rand(*args)*2-1)*scl

def interp2D(a, b, u):
    """Linearly interpolate by amount u from point a to point b."""
    return point2D(a[0]+u*(b[0]-a[0]), a[1]+u*(b[1]-a[1]))

######################################################################

class Interval1D:
    """Class to represent a closed interval on the real number line."""

    def __init__(self, lower=None, upper=None):
        """Initialize with given lower and upper bound. If
        lower>upper, the interval is assumed to be empty."""
        if lower is None and upper is None:
            self.lower = 1.0
            self.upper = 0.0
        elif upper is None:
            self.lower = lower
            self.upper = lower
        else:
            self.lower = lower
            self.upper = upper

    def __str__(self):
        """ Stringify this object. """
        return '[{0:g}, {1:g}]'.format(self.lower, self.upper)
    
    def copy(self):
        """Return a copy of this interval."""
        return Interval1D(self.lower, self.upper)

    def empty(self):
        """The interval is empty if its lower bound is greater than
        its upper bound"""
        return self.lower > self.upper

    def contains(self, value, d=0.0):
        """Return true if the given value lies with a distance of d
        outside the interval."""
        if self.upper < self.lower:
            return False
        else:
            return self.lower-d <= value and self.upper+d >= value

    def extend(self, value):
        """Enlarge the interval so that the given value is included inside it."""
        if self.lower > self.upper:
            self.lower = value
            self.upper = value
        else:
            self.lower = min(self.lower, value)
            self.upper = max(self.upper, value)
        
    def overlaps(self, that, d=0.0):
        """Returns true if this interval is within a distance d of
        overlapping the given interval."""
        return ( self.lower <= self.upper and
                 that.lower <= that.upper and
                 self.lower-d <= that.upper and
                 self.upper+d >= that.lower )

    def hull(self, that):
        """Returns the smallest interval containing both intervals given."""
        if (self.lower > self.upper):
            return Interval1D(that.lower, that.upper)
        elif (that.lower > that.upper):
            return Interval1D(self.lower, self.upper)
        else:
            return Interval1D(min(self.lower, that.lower), 
                              max(self.upper, that.upper))

    def intersection(self, that):
        """Returns the intersection of the two intervals given, or the
        empty interval if they do not overlap."""
        if (self.lower > self.upper) or (that.lower > that.upper):
            return Interval1D(1.0, 0.0)
        else:
            return Interval1D(max(self.lower, that.lower), 
                              min(self.upper, that.upper))

    def length(self):
        """Returns the length of the interval, or zero if empty."""
        return max(self.upper-self.lower, 0.0)

    def wrap(self, value):
        """Returns the given value, wrapped into this range."""
        return wrap(value, self.lower, self.upper)

    def clamp(self, value):
        """Returns the given value, wrapped into this range."""
        return clamp(value, self.lower, self.upper)

    def dilate(self, d):
        """Expands this interval by the given value on both sides, if
        non-empty."""
        if self.upper >= self.lower:
            self.upper += d
            self.lower -= d

    def dilated(self, d):
        """Returns a copy of this interval expanded by the given value
        on both sides, if non-empty."""
        if self.upper >= self.lower:
            return Interval1D(self.lower-d, self.upper+d)
        else:
            return Interval1D(self.lower, self.upper)

    def random(self, *args):
        """Return a random value in [lower, upper)."""
        return numpy.random.rand(*args)*self.length() + self.lower

######################################################################

class Box2D:

    """Class to implement a 2D Box as closed intervals in x and y."""

    def __init__(self, *args):
        """Init can be called in multiple ways: with four parameters
        (x0, y0, x1, y1), with two intervals (xinterval, yinterval),
        or with no arguments, which creates an empty box."""
        if len(args) == 4:
            (x0, y0, x1, y1) = args
            self.xinterval = Interval1D(x0,x1)
            self.yinterval = Interval1D(y0,y1)
        elif len(args) == 2:
            (self.xinterval, self.yinterval) = args
        elif len(args) == 0:
            self.xinterval = Interval1D()
            self.yinterval = Interval1D()
        else:
            raise Exception('invalid initializer for Box2D')

    def __str__(self):
        """Stringify this object."""
        return str(self.xinterval) + 'x' + str(self.yinterval)

    def copy(self):
        """Return a copy of this object."""
        return Box2d(self.xinterval.copy(), self.yinterval.copy())

    def empty(self):
        """Return true if either the x or y interval is empty."""
        return self.xinterval.empty() or self.yinterval.empty()

    def pmin(self):
        """Return the (x,y) pair corresponding to the lower bounds of
        the x and y intervals."""
        return point2D(self.xinterval.lower, self.yinterval.lower)

    def pmax(self):
        """Return the (x,y) pair corresponding to the upper bounds of
        the x and y intervals."""
        return point2D(self.xinterval.upper, self.yinterval.upper)

    def dims(self):
        """Return the (x,y) pair corresponding to the lengths of the
        the x and y intervals."""
        return point2D(self.xinterval.length(), self.yinterval.length())

    def center(self):
        """Return the (x,y) pair corresponding to the midpoints of the
        the x and y intervals."""
        return point2D(0.5*(self.xinterval.lower + self.xinterval.upper),
                       0.5*(self.yinterval.lower + self.yinterval.upper))

    def dilate(self, d):
        """Enlarges this box by the amount d on each side."""
        self.xinterval.dilate(d)
        self.yinterval.dilate(d)

    def dilated(self, d):
        """Returns a copy of this box enlarged by the amount d on each side."""
        return Box2D(self.xinterval.dilated(d), self.yinterval.dilated(d))

    def extend(self, pt):
        """Enlarges this box to include the (x,y) pair given in pt."""
        self.xinterval.extend(pt[0])
        self.yinterval.extend(pt[1])
    
    def contains(self, pt, d=0.0):
        """Returns true if the (x,y) pair given in pt lies within a
        distance d outside the box."""
        return ( self.xinterval.contains(pt[0], d) and 
                 self.yinterval.contains(pt[1], d) )
    
    def overlaps(self, that, d=0.0):
        """Returns true if this box overlaps the given box."""
        return ( self.xinterval.overlaps(that.xinterval, d) and
                 self.yinterval.overlaps(that.yinterval, d) )


    def hull(self, that):
        """Returns the smallest box containing both this box and the one given."""
        return Box2D( self.xinterval.hull(that.xinterval),
                      self.yinterval.hull(that.yinterval) )

    def intersection(self, that):
        """Returns the intersection of this box and the one given."""
        return Box2D( self.xinterval.intersection(that.xinterval),
                      self.yinterval.intersection(that.yinterval) )

    def wrap(self, point):
        """Return the point wrapped into the x- and y-ranges."""
        return point2D(self.xinterval.wrap(point[0]),
                       self.yinterval.wrap(point[1]))

    def clamp(self, point):
        """Return the point clamped into the x- and y-ranges."""
        return point2D(self.xinterval.clamp(point[0]),
                       self.yinterval.clamp(point[1]))

    def random(self):
        """Return a random point inside the box."""
        return point2D(self.xinterval.random(), self.yinterval.random())

######################################################################

class Transform2D:
    
    """ Class to implement simple 2D transformations, with points
    represented as 2D numpy arrays."""
    def __init__(self, *args):
        """ Initialize as Transform2D(x,y,theta), or
        Transform2D((x,y),theta) or Transform2D((x,y)) or
        Transform2D(theta) or Transform2D() (identity)."""
        if len(args) == 3:
            (self.pos, self.theta) = (point2D(args[0], args[1]), args[2])
        elif len(args) == 2:
            if isinstance(args[0], collections.Iterable):
                self.pos = point2D(args[0])
                self.theta = args[1]
            else:
                self.pos = point2D(args)
                self.theta = 0.0
        elif len(args) == 1:
            if isinstance(args[0], collections.Iterable):
                self.pos = point2D(args[0])
                self.theta = 0.0
            else:
                self.pos = point2D(0.0, 0.0)
                self.theta = args[0]
        elif len(args) == 0:
            (self.pos, self.theta) = (point2D(0.0, 0.0), 0.0)
        else:
            raise Exception('invalid initializer for Transform2D')


    def __str__(self):
        """Stringify this object."""
        return '(x:{0:g}, y:{1:g}, theta:{2:g})'.format(
            self.pos[0], self.pos[1], self.theta)

    def copy(self):
        """Return a copy of this object."""
        return Transform2D(point2D(self.pos), self.theta)

    def transformFwd(self, point):
        """Return the forward transformation of the given point:
        output = R(theta) * point + (x, y)"""
        (ct, st) = point2DFromAngle(self.theta)
        return point2D(ct*point[0] - st*point[1] + self.pos[0], 
                       st*point[0] + ct*point[1] + self.pos[1])
    
    def transformInv(self, point):
        """Return the inverse transformation of the given point:
        output = R(-theta) * [point - (x,y)]"""
        (ct, st) = point2DFromAngle(self.theta)
        xx = point[0] - self.pos[0]
        yy = point[1] - self.pos[1]
        return point2D(ct*xx + st*yy, -st*xx + ct*yy)

    def rotFwd(self, point):
        """ Return the forward rotation-only transformation of the
        given point: output = R(theta) * point"""
        (ct, st) = point2DFromAngle(self.theta)
        return point2D(ct*point[0] - st*point[1], 
                       st*point[0] + ct*point[1])
    
    def rotInv(self, point):
        """ Return the inverse rotation-only transformation of the
        given point: output = R(-theta) * point"""
        (ct, st) = point2DFromAngle(self.theta)
        return point2D(ct*point[0] + st*point[1], -st*point[0] + ct*point[1])

    def composeWith(self, other):
        """Return the product of this transformation with other such
        that the transforming a point forward via the result is
        equivalent to first transforming it forward via the other,
        then with this transformation."""
        (nx, ny) = self.transformFwd( other.pos )
        return Transform2D(nx, ny, self.theta + other.theta)

    def distTo(self, other):
        """Return a (xydist, angledist) tuple indicating the
        displacement of the two transforms, where angledist is a
        signed distance."""
        return (dist2D(self.pos, other.pos), 
                angleDiff(self.theta, other.theta))

    def absDistTo(self, other):
        """Return a (xydist, angledist) tuple indicating the
        displacement of the two transforms, where angledist is an
        unsigned distance."""
        return (dist2D(self.pos, other.pos), 
                math.fabs(angleDiff(self.theta, other.theta)))

    def interpTo(self, other, u):
        """Return the transformation obtained by linearly
        interpolating the translation and rotations of this transform
        and the given transform by an amount u."""
        pdiff = other.pos - self.pos
        adiff = angleDiff(self.theta, other.theta)
        return Transform2D(self.pos + u*pdiff, self.theta + u*adiff)

    def inverse(self):
        """ Return the transformation which undoes this transformation. """
        (xx,yy) = self.transformInv((0,0))
        return Transform2D(xx,yy,-self.theta)

    def __mul__(self, obj):
        """ Overloaded multiplication. If the right operand is a
        Transform2D, returns self.composeWith(obj), otherwise returns
        self.transformFwd(obj). """
        if isinstance(obj, Transform2D):
            return self.composeWith(obj)
        else:
            return self.transformFwd(obj)


######################################################################

def fixPolygon2DPoints(pts):

    """Return an Nx2 numpy array of type float32, one point per row."""

    # copy data
    rval = numpy.array(pts,dtype='float32')
    
    if len(rval.shape) == 1:
        rval = rval.reshape(-1,2)
    elif rval.shape[1] != 2 and rval.shape[0] == 2:
        rval = rval.transpose()

    return rval


######################################################################

class Polygon2D:

    """Class to encapsulate a 2D Polygon. The polygon is assumed to be
    simple (no self-intersections) and convex, and some methods may
    behave unexpectedly if either assumption is violated."""

    def __init__(self, data):

        """Initialize this polygon. Data can be an array of
        coordinates, or also a Box2D."""

        if isinstance(data, Box2D):
            (x0,y0) = data.pmin()
            (x1,y1) = data.pmax()
            data = [x1,y1,x0,y1,x0,y0,x1,y0]

        self.points = fixPolygon2DPoints(data)

        self.updatePoints()

    def __str__(self):
        """Return a string representation of this object."""
        return 'Polygon2D:\n'+str(self.points)

    def copy(self):
        """Return a copoy of this object."""
        return Polygon2D(self.points)

    def draw(self, filled=True):
        """Draw using OpenGL by creating a Vertex Buffer Object (VBO)."""
        if self.vvbo is None:
            idx = []

            for i in range(self.npoints-2):
                idx.append(0)
                idx.append(i+1)
                idx.append(i+2)

            self.glpoints = self.points
            self.glindices = numpy.array(idx, dtype='uint16')

            self.vvbo = vbo.VBO(self.glpoints, usage=gl.GL_STATIC_DRAW)
            self.ivbo = vbo.VBO(self.glindices, usage=gl.GL_STATIC_DRAW, 
                                    target=gl.GL_ELEMENT_ARRAY_BUFFER)

            gl.glEnableClientState(gl.GL_VERTEX_ARRAY)

        self.vvbo.bind()

        if filled:

            self.ivbo.bind()

            gl.glVertexPointer(2, gl.GL_FLOAT, 0, self.vvbo)
            
            gl.glDrawElements(gl.GL_TRIANGLES, self.glindices.shape[0], 
                              gl.GL_UNSIGNED_SHORT, self.ivbo)

            self.ivbo.unbind()

        else:
            gl.glVertexPointer(2, gl.GL_FLOAT, 0, self.vvbo)
            gl.glDrawArrays(gl.GL_LINE_LOOP, 0, self.glpoints.shape[0])
            

        self.vvbo.unbind()

    def updatePoints(self):
        """This should be called whenever the points are modified (and
        is called automatically by the initializer). It computes the
        area and centroid of the polygon, as well as a list of
        directions and projections used for collision detection."""
        self.npoints = self.points.shape[0]

        self.vvbo = None
        self.ivbo = None

        self.area = 0
        self.centroid = point2D(0,0)
        self.directions = numpy.zeros((self.npoints,2),dtype='f')
        self.dintervals = []
        ndirs = 0

        for i in range(self.npoints):
            ii = (i+self.npoints-1) % self.npoints
            p1 = self.points[ii]
            p2 = self.points[i]
            z = crossProduct2D(p1, p2)
            self.area += z
            self.centroid += (p1+p2)*z
            di = normal2D(p1, p2)
            if not -di in self.directions[0:ndirs,:]:
                self.directions[ndirs,:] = di
                ndirs += 1
                self.dintervals.append(self.project(di))

        self.directions = self.directions[0:ndirs,:]

        self.area *= 0.5
        self.centroid *= 1.0/(6*self.area)

        self.radius = 0

        for i in range(self.npoints):
            self.radius = max(self.radius, 
                              distSqr2D(self.points[i], self.centroid))

        self.radius = math.sqrt(self.radius)


    def project(self, axis):
        """Returns the extents of this polygon when projected onto the
        given axis. Used for collision detection."""
        proj = self.points[:,0]*axis[0] + self.points[:,1]*axis[1]
        return Interval1D(proj.min(), proj.max())

    def closestDistSqr(self, pt):
        """Return the square of the distance from the given point to
        the closest point in the polygon. Used for collision
        detection."""
        dists = (self.points[:,0]-pt[0])**2 + (self.points[:,1]-pt[1])**2
        return dists.min()

    def contains(self, point, d=0.0):

        """Returns true if the given point is within a distance d of
        the interior of this polygon."""

        distSqr = distSqr2D(self.centroid, point)
        minDist = self.radius + d

        if distSqr > minDist*minDist:
            return False

        inside = True
        for i in range(self.npoints):
            ii = (i+self.npoints-1) % self.npoints
            p1 = self.points[ii]
            p2 = self.points[i]
            pdiff = point - p1
            ldiff = p2 - p1
            u = dot2D(pdiff,ldiff)/normSqr2D(ldiff)
            ldist = dot2D(pdiff, normal2D(p1,p2))
            if ldist > d:
                return False
            elif ldist > 0:
                if u>=0 and u<=1:
                    return True
                else:
                    inside = False


        return inside or self.closestDistSqr(point) < d*d

    def overlapsLine(self, p1, p2, d=0.0):

        """Returns true if the line segment given by p1, p2 is within
        a distance d of overlapping this polygon."""

        line_axis = normal2D(p1,p2)

        pdiff = p1-self.centroid
        ldist = math.fabs(dot2D(pdiff, line_axis))

        if ldist > self.radius + d:
            return False

        iself = self.project(line_axis)
        k = dot2D(p1,line_axis)
        iline = Interval1D(k,k)
        if not iself.overlaps(iline, d):
            return False

        for i in range(len(self.directions)):
            self_axis = self.directions[i]
            iself = self.dintervals[i]
            l1 = dot2D(p1, self_axis)
            l2 = dot2D(p2, self_axis)
            iline = Interval1D(min(l1,l2), max(l1,l2))
            if not iself.overlaps(iline, d):
                return False

        return True

######################################################################

def _collidesPolygonPolygon(pa, ta, pb, tb, d=0.0):

    """Utility function to implement polygon-polygon collisions. Do
    not use this method: rather, use collidesPolygonPolygon (with no
    leading underscore)."""

    w_from_a = ta
    w_from_b = tb

    a_from_b = w_from_a.inverse() * w_from_b

    for i in range(pa.directions.shape[0]):

        a_axis = pa.directions[i]
        ia = pa.dintervals[i]

        b_axis = a_from_b.rotInv(a_axis)
        offset = dot2D(a_from_b.pos, a_axis)

        ib = pb.project(b_axis)

        ib.lower += offset
        ib.upper += offset

        if not ia.overlaps(ib, d):
            return False

    return True

######################################################################

def collidesPolygonPolygon(pa, ta, pb, tb, d=0.0):

    """Given two polygons pa and pb, and two transformations ta and tb
    which transform points from polygon frame to world frame, returns
    true if the polygons are within a distance d of overlapping."""

    # bounding circle test
    ca = ta * pa.centroid
    cb = tb * pb.centroid

    distSqr = distSqr2D(ca, cb)

    minDist = pa.radius + pb.radius + d

    if distSqr > minDist*minDist:
        return False

    if not _collidesPolygonPolygon(pa, ta, pb, tb, d):
        return False
    elif not _collidesPolygonPolygon(pb, tb, pa, ta, d):
        return False
    else:
        return True

######################################################################

def collidesPolygonLine(pa, ta, l1, l2, d=0.0):

    """Given a polygon pa and a rigid transformation ta mapping points
    from the polygon frame to the world frame, returns true if the
    line segment from l1 to l2 is within a distance of d of
    overlapping the polygon."""
    return pa.overlapsLine(ta.transformInv(l1), 
                           ta.transformInv(l2), d)

######################################################################

def collidesPolygonPoint(pa, ta, p, d=0.0):

    """Given a polygon pa and a rigid transformation ta mapping points
    from the polygon frame to the world frame, returns true if the
    given point is within a distance d of the interior of the
    polygon."""

    return pa.contains(ta.transformInv(p), d)

######################################################################

class Shape2D:

    """Class for encapsulating a polygon and a rigid transform, as
    well as an optional color for drawing."""

    def __init__(self, polygon, xform=None, color=None, name=None):

        """Initialize this object. The polygon argument may be an
        instance of the Polygon2D class, or else an object that can be
        passed to the Polygon2D initializer."""
        
        if isinstance(polygon, Polygon2D):
            self.polygon = polygon
        else:
            self.polygon = Polygon2D(polygon)

        if color is None:
            self.color = numpy.array((0.75,0.75,0.75))
        else:
            self.color = numpy.array(color)

        if isinstance(xform, Transform2D):
            self.xform = xform
        elif xform is None:
            self.xform = Transform2D()
        else:
            self.xform = Transform2D(*xform)

        if name is None:
            self.name = 'Shape2D'
        else:
            self.name = name

    def copy(self):
        """Return a copy of this object."""
        return Shape2D(self.polygon.copy(), 
                       xform=self.xform.copy(), 
                       color=numpy.array(color), 
                       name=self.name)

    def draw(self, colored=True, filled=True):

        """Draw this shape using OpenGL."""

        if colored:
            gl.glColor(self.color)

        (ct,st) = point2DFromAngle(self.xform.theta)
        (x,y) = (self.xform.pos)
        
        self.glmatrix = numpy.array([ct, st, 0, 0, 
                                     -st, ct, 0, 0,
                                     0, 0, 1, 0, 
                                     x, y, 0, 1], dtype='f').reshape(4,-1)

        gl.glPushMatrix()
        gl.glMultMatrixf(self.glmatrix)

        self.polygon.draw(filled=filled)

        gl.glPopMatrix()

    def project(self, axis):
        """Returns the extents of this shape when projected onto the
        given axis. Can be used for collision detection or determining
        supporting hulls."""
        rot_axis = self.xform.rotInv(axis)
        offset = dot2D(axis, self.xform.pos)
        interval = self.polygon.project(rot_axis)
        interval.lower += offset
        interval.upper += offset
        return interval

    def overlaps(self, that, d=0.0):
        """Returns true if this shape is within a distance d of
        overlapping the given shape."""
        return collidesPolygonPolygon(self.polygon, self.xform,
                                      that.polygon, that.xform, d)

    def contains(self, point, d=0.0):
        """Returns true if the given point is within a distance d of
        the interior of this shape."""
        return collidesPolygonPoint(self.polygon, self.xform, point, d)

    def overlapsLine(self, p1, p2, d=0.0):
        """Returns true if the line segment given by p1, p2 is within
        a distance d of overlapping this shape."""
        return collidesPolygonLine(self.polygon, self.xform, p1, p2, d)


