import OpenGL.GL as gl
import OpenGL.GLU as glu
import OpenGL.GLUT as glut

import Image

import os
import os.path

import signal
import sys
import threading
import numpy
import math


import GeomUtil as gutil

"""Module to encapsulate some basic functionality of GLUT using Python
classes. Written for E91 Humanoid Robotics, Spring 2013, by Matt
Zucker."""

GLUTAPP_DEFAULT_MODE = ( glut.GLUT_DOUBLE | glut.GLUT_RGB | 
                         glut.GLUT_DEPTH | glut.GLUT_MULTISAMPLE )

class GlutApp:

    """Class to encapsulate a single-window GLUT
    application. Applications can be created relatively easily by
    inheriting from this class."""

    def __init__(self, title="GlutApp", 
                 mode=GLUTAPP_DEFAULT_MODE, winsz=(640,480)):

        """Initialize this object by giving window title, display mode
        (default is double-buffered RGB with depth buffer and
        multisampling), and window size as tuple or list."""


        glut.glutInit()
        glut.glutInitWindowSize(winsz[0], winsz[1])
        glut.glutInitDisplayMode(mode)

        glut.glutCreateWindow(title)

        glut.glutDisplayFunc(self.display)
        glut.glutReshapeFunc(self.reshape)
        glut.glutKeyboardFunc(self.keyboard)
        glut.glutSpecialFunc(self.special)
        glut.glutMouseFunc(self.mouse)
        glut.glutMotionFunc(self.motion)

        self._circlePolygon = None

        self.initGL()
        signal.signal(signal.SIGINT, self._ctrlC)
        glut.glutTimerFunc(100, self._watchdog, 0)


    def _watchdog(self, value):
        """This function is registered as a timer callback to get
        Ctrl-C working from the command line. You should not need to
        call it."""
        glut.glutTimerFunc(100, self._watchdog, 0)

    def _ctrlC(self, signum, frame):
        """This function is registered as a signal handler to get
        Ctrl-C working from the command line. You should not need to
        call it."""
        print 'Ctrl-C pressed, exiting.'
        sys.exit(0)

    def setTimer(self, delay, value=0):
        """Register a callback to self.timer after delay
        milliseconds. The given integer value (default zero) will be
        passed to the callback function. This is a "single-shot" timer
        -- to get it to repeat, you will likely want to call setTimer
        again inside of the self.timer callback."""
        glut.glutTimerFunc(delay, self.timer, value)

    def timer(self, value):
        """Callback registered by self.setTimer(). You can override
        this method in your own subclass of GlutApp."""
        pass


    def mouse(self, button, state, x, y):
        """Mouse callback called whenever a mouse button is pressed or
        released. Left button corresponds to GLUT_LEFT, etc. State is
        GLUT_DOWN or GLUT_UP, and x, y are window coordinates. You can
        redefine this method in a subclass to support mouse events."""
        pass

    def motion(self, x, y):
        """Mouse callback called whenever the mouse is moved while a
        button is pressed. Window coordinates are given in x and
        y. You can redefine this method in a subclass to support mouse
        events."""
        pass

    def run(self):
        """Run the application after initialization."""
        glut.glutMainLoop()

    def keyboard(self, key, x, y):
        """Keyboard callback called whenever a key is pressed. The key
        is given as a single-character string. Default implementation
        simply exits Python when ESC (character 27) is pressed. You
        can redefine this method in a subclass to support keyboard
        events."""
        if (key == '\x1b'):
            print 'ESC pressed, exiting.'
            sys.exit(0)


    def special(self, key, x, y):
        """Keyboard callback called whenever a special key (function,
        arrows, etc) is pressed. The key is given as numeric constant,
        see GLUT documentation for details. You can redefine this
        method in a subclass to support keyboard events."""
        pass


    def initGL(self):

        """This function is called when the app is initialized. The
        default implementation sets some basic parameters for OpenGL
        like point size, line width, and screen color."""

        gl.glClearColor(1.0,1.0,1.0,0.0)

        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()

        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()

        gl.glPointSize(5.0)
        gl.glLineWidth(2.0)


    def reshape(self, w, h):

        """Reshape callback called whenever the window is resized. You
        will likely want to redefine this method in a subclass to do
        application-specific setup of the OpenGL viewport and
        projection matrix."""
        
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glViewport(0,0,w,h)

        gl.glMatrixMode(gl.GL_MODELVIEW)


    def display(self):

        """Display callback called whenever the screen needs to be
        repainted. To trigger this callback, you can also call
        OpenGL.GLUT.glutPostRedisplay() from the GLUT module. You will
        likely want to redefine this method in a subclass to draw your
        scene."""

        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()

        glut.glutSwapBuffers()

    def setupOrtho2D(self, w, h, box):

        """Set up the OpenGL projection matrix to define an orthogonal
        2D projection for the scene with given window dimensions (w,h)
        and scene bounds given in the GeomUtil.Box2D object box. This
        method ensures that the projection contains the box given,
        with the correct aspect ratio."""

        (bw, bh) = box.dims()
        (cx, cy) = box.center()

        ba = float(bw)/bh
        wa = float(w)/h
        
        if (ba > wa):
            bh *= ba/wa
        else:
            bw *= wa/ba
            
        gl.glViewport(0, 0, w, h)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        glu.gluOrtho2D(cx - 0.5*bw, cx + 0.5*bw,
                   cy - 0.5*bh, cy + 0.5*bh)

        gl.glMatrixMode(gl.GL_MODELVIEW)

        self.pMatrix = gl.glGetDoublev(gl.GL_PROJECTION_MATRIX)
        self.mMatrix = gl.glGetDoublev(gl.GL_MODELVIEW_MATRIX)
        self.viewport = gl.glGetIntegerv(gl.GL_VIEWPORT)

    def unproject2D(self, x, y):

        """Return a 2D point corresponding to the given x,y window
        coordinates (e.g. from a mouse or keyboard callback) obtained
        by inverting the OpenGL projection and modelview matrices."""

        glcoords = glu.gluUnProject(x, self.viewport[3]-y, 0, 
                                    self.mMatrix, self.pMatrix, self.viewport)

        return gutil.point2D(glcoords[0:2])

    def drawCircle(self, point, rad, slices=32, filled=True):

        """Draw a filled 2D circle at the given (x,y) point with the
        given radius. Slices determines the number of angular
        subdivisions (default 32)."""

        gl.glPushMatrix()

        gl.glTranslate(point[0], point[1], 0)
        gl.glScale(rad, rad, rad)

        if not self._circlePolygon or self._circlePolygon.npoints != slices:
            angles = 2*math.pi*numpy.arange(slices)/slices
            points = [ gutil.point2DFromAngle(x) for x in angles ]
            self._circlePolygon = gutil.Polygon2D(points)

        self._circlePolygon.draw(filled=filled)

        gl.glPopMatrix()

    def drawBitmapString(self, str, font=glut.GLUT_BITMAP_8_BY_13):
        
        """Draw the text given in str at the current raster position
        and raster color. To set the position and color, use the
        something like the following code:

          glColor3f(0,0,0)
          glRasterPos(0,0)
          self.drawBitmapString("hi world!")

        You must set the color before the raster position in order to
        take effect properly."""

        for char in str:
            glut.glutBitmapCharacter(font, ord(char))

class FrameGrabber:
    """Class to encapsulate simple frame grabbing from OpenGL to a
    sequence files with a regular pattern in a separate directory."""

    def __init__(self, outputDir=None, pattern=None):
        """Initialize this object. Simply calls setup(outputDir,
        pattern) if they are provided."""
        if outputDir is not None and pattern is not None:
            self.setup(outputDir, pattern)

    def setup(self, outputDir, pattern):
        """Set the output directory and pattern for the frames. The
        filename is generated by

           os.path.join(self.outputDir, 
                        self.pattern.format(self.curIndex))

        so if self.outputDir is 'frames' and self.format is equal to
        'image{:04d}.png' and self.currentIndex is equal to 3, the
        resulting filename would be 'frames/image0003.png'.  This
        method automatically sets self.currentIndex to zero, but you
        can change it yourself if you want to start at a different
        number."""
        self.outputDir = outputDir
        self.pattern = pattern

        self.curIndex = 0

        if not os.path.exists(outputDir):
            os.makedirs(outputDir)

    def grabFrame(self, verbose=False):
        """Grabs a frame from the current OpenGL context and dump it
        to the next filename in the sequence. If verbose is True,
        prints a line of output to the console."""
        (w,h) = gl.glGetIntegerv(gl.GL_VIEWPORT)[2:4]

        gl.glFlush()
        buf = gl.glReadPixels(0, 0, w, h, gl.GL_RGB, gl.GL_UNSIGNED_BYTE)

        img = Image.fromstring(mode='RGB', size=(w,h), data=buf)
        img = img.transpose(Image.FLIP_TOP_BOTTOM)

        outputFile = os.path.join(self.outputDir, 
                                  self.pattern.format(self.curIndex))

        img.save(outputFile)

        if verbose:
            print 'wrote ' + outputFile

        self.curIndex += 1


class DragManager:

    """Class to encapsulate simple 2D shape dragging operations for
    OpenGL using the GeomUtil.Shape2D class and the GlutApp.Draggable
    helper class."""

    def __init__(self, autoPost=True):

        """Initialize. Set autoPost to False to prevent this object
        from calling glutPostRedisplay() when mouse is clicked, moved,
        etc."""

        self.draggables = []
        self.current = None
        self.autoPost = autoPost

    def isDragging(self):

        """Returns true if an object is currently being dragged."""
        return self.current is not None

    def add(self, obj):
        """Add the given object to the list of managed objects. It
        should either be an instance of gutil.Shape2D or a
        GlutApp.Draggable object. The list of objects can be accessed
        via self.draggables."""

        if isinstance(obj, gutil.Shape2D):
            obj = Draggable(obj)
        elif not isinstance(obj, Draggable):
            raise Exception('non-draggable added in DragManager.addDraggable')
        self.draggables.append(obj)
        return obj

    def remove(self, d):
        """Remove the given object from the list of managed
        objects."""
        self.draggables.remove(d)
        if d is self.current:
            self.current = None

    def mouseDown(self, point):
        """Call this when the mouse is clicked to pass the click on to
        the list of managed objects. The first object in the list
        containing the given (x,y) point will get its own mouseDown
        method called. The point given should be expressed in scene,
        not window coordinates."""
        self.current = None
        for d in self.draggables:
            if d.contains(point):
                self.current = d
                self.current.mouseDown(point)
                if self.autoPost:
                    glut.glutPostRedisplay()
                break

    def mouseUp(self, point):
        """Call this when the mouse is released to pass the message to
        the currently selected object. The point given should be
        expressed in scene, not window coordinates."""
        if self.current is not None:
            self.current.mouseUp(point)
            if self.autoPost:
                glut.glutPostRedisplay()
        self.current = None

    def mouseMoved(self, point):
        """Call this when the mouse is moved to pass the message to
        the currently selected object. The point given should be expressed in
        scene, not window coordinates."""
        if self.current is not None:
            self.current.mouseMoved(point)
            if self.autoPost:
                glut.glutPostRedisplay()

class Draggable:

    """Class to help drag GeomUtil.Shape2D objects by modifying their
    transformation via mouse interaction."""

    def __init__(self, shape, clickTol=0.0, 
                 onMouseDown=None,
                 onMouseMoved=None,
                 onMouseUp=None):

        """Initialize to start dragging the given shape. You can
        provide extra notification functions if you want to receive
        callbacks when the object is interacted with, or you can also
        subclass this object and override the onMouseDown,
        onMouseMoved, and onMouseUp methods instead."""

        self.shape = shape
        self.clickTol = clickTol
        self.mouseDownPoint = None
        self.mouseDownCallback = onMouseDown
        self.mouseMovedCallback = onMouseMoved
        self.mouseUpCallback = onMouseUp

    def onMouseDown(self, point):
        """Called when the mouse is clicked on this object. Default
        implementation is a no-op."""
        pass

    def onMouseMoved(self, point):
        """Called when the this object is moved when the mouse is
        pressed. Default implementation is a no-op."""
        pass

    def onMouseUp(self, point):
        """Called when the this object is released by the
        mouse. Default implementation is a no-op."""
        pass

    def contains(self, point):
        """Determine whether the shape managed by this object contains
        the given point."""
        return self.shape.contains(point, self.clickTol)

    def mouseDown(self, point):
        """Method called by DragManager to notify that dragging has
        started."""
        self.mouseDownPoint = point
        self.lastMousePoint = point
        if self.mouseDownCallback is not None:
            self.mouseDownCallback(point)
        else:
            self.onMouseDown(point)

    def mouseMoved(self, point):
        """Method called by DragManager to notify that mouse has moved
        while dragging."""
        self.shape.xform.pos += point - self.lastMousePoint
        self.lastMousePoint = point
        if self.mouseMovedCallback is not None:
            self.mouseMovedCallback(point)
        else:
            self.onMouseMoved(point)
        
    def mouseUp(self, point):
        """Method called by DragManager to notify that dragging has
        ended."""
        self.mouseDownPoint = None
        self.lastMousePoint = None
        if self.mouseUpCallback is not None:
            self.mouseUpCallback(point)
        else:
            self.onMouseUp(point)
