import GeomUtil as gutil
import numpy
import math
import OpenGL.GL as gl

"""Set of classes to manage 2D kinematic models, loosely based on
OpenRAVE. Written for E91 Humanoid Robotics, Spring 2013, by Matt
Zucker."""

################################################################################

class Body:

    """Class to encapsulate a single rigid body in a kinematic
    model. A body may have a parent body (self.offsetFrom), has an
    associated 2D transform (self.xform), and a list of shapes
    (self.shapes) associated with it (which may have their own
    associated transforms."""
    
    def __init__(self, name, 
                 offsetFrom=None, 
                 xform=gutil.Transform2D(),
                 mass=0.0,
                 com=gutil.point2D(0,0),
                 shapes=[]): 

        """Initialize this Body."""

        self.name = name
        self.index = -1
        self.offsetFrom = offsetFrom
        self.xform = xform
        self.parentJoint = None
        self.childJoints = []
        self.childBodies = []
        self.mass = mass
        self.com = com
        self.shapes = shapes

################################################################################

class Joint:

    """Class to encapsulate a hinge joint in a kinematic model. A
    joint has a parent body (self.body1) and a child body
    (self.body2), and it's anchor position may be expressed relative
    to either one of them (self.offsetFrom, self.anchor). It may
    optionally also have joint limits (self.limits) expressed as a
    GeomUtil.Interval1D object. The default limits are the empty
    interval (indicating no joint limits)."""

    def __init__(self, name,
                 body1,
                 body2,
                 offsetFrom,
                 anchor,
                 limits=gutil.Interval1D()):

        """Initialize this Joint."""

        self.name = name
        self.index = -1
        self.body1 = body1
        self.body2 = body2
        self.offsetFrom = offsetFrom
        self.anchor = anchor
        self.limits = limits

################################################################################

class Manipulator:

    """Class to encapsulate an end effector in a kinematic model. A
    Manipulator is attached to a body (self.effectorBody), and has an
    associated transform (self.xform) relative to the body's
    coordinate frame. It is aware of the joints (self.joints) tracing
    the manipulator back to the root body of the tree
    (self.baseBody)."""

    def __init__(self, name,
                 effectorBody,
                 xform,
                 baseBody=None):

        """Initalize this Manipulator."""

        self.name = name
        self.index = -1
        self.effectorBody = effectorBody
        self.xform = xform
        self.baseBody = baseBody
        self.joints = []
        self.l1 = None
        self.l2 = None
        self.a1 = None
        self.a2 = None
        self.initJointPos = None
        self.activeBodies = {}

################################################################################

KINMODEL_IK_IGNORE = 0
KINMODEL_IK_BPOS   = 1
KINMODEL_IK_BXFORM = 2
KINMODEL_IK_WPOS   = 3
KINMODEL_IK_WXFORM = 4

################################################################################

class KinModel:

    """Class to encapsulate a full 2D kinematic tree model consisting
    of Body, Joint, and Manipulator objects. You can access them via
    self.bodies, self.joints, and self.manipulators."""

    def __init__(self, name):

        """Initialize this KinModel object."""

        self.name = name
        self.bodies = []
        self.joints = []
        self.manipulators = []
        self.ikPosTol = 1e-5
        self.ikRotTol = 1e-4
        self._bodyDependsOnJoint = None
        self._isAncestorOf = None
        
    def add(self, obj):
        """Add the given object to this model. The object should be an
        instance of the Body, Joint, or Manipulator classes."""
        if isinstance(obj, Body):
            return self.addBody(obj)
        elif isinstance(obj, Joint):
            return self.addJoint(obj)
        elif isinstance(obj, Manipulator):
            return self.addManipulator(obj)
        else:
            raise Exception('unrecognized object type in KinModel.add')

    def addBody(self, body):
        """Add the given Body to this model."""
        assert( not body in self.bodies )
        assert( body.offsetFrom is None or body.offsetFrom in self.bodies )
        body.index = len(self.bodies)
        if body.offsetFrom:
            body.offsetFrom.childBodies.append(body)
        self.bodies.append(body)
        return body
    
    def addJoint(self, joint):
        """Add the given Joint to this model."""
        assert( not joint in self.joints )
        assert( joint.body1 in self.bodies )
        assert( joint.body2 in self.bodies )
        assert( joint.body2.parentJoint is None )
        assert( joint.body1 is joint.offsetFrom or
                joint.body2 is joint.offsetFrom )
        assert( joint.body2.offsetFrom is joint.body1 )
        assert( len(joint.anchor) == 2 )
        assert( isinstance(joint.limits, gutil.Interval1D) )
        joint.index = len(self.joints)
        joint.body2.parentJoint = joint
        joint.body1.childJoints.append(joint)
        self.joints.append(joint)
        return joint

    def addManipulator(self, manip):
        """Add the given Manipulator to this model."""
        assert( not manip in self.manipulators )
        assert( manip.effectorBody in self.bodies )
        assert( manip.baseBody is None or manip.baseBody in self.bodies )

        for j in manip.joints:
            assert( j in self.joints )

        for b in manip.activeBodies.keys():
            assert( b in self.bodies )

        if not manip.joints:

            body = manip.effectorBody

            while body.parentJoint or body.offsetFrom:
                if body.parentJoint:
                    manip.joints.append(body.parentJoint)
                    body = body.parentJoint.body1
                else:
                    body = body.offsetFrom
                if body is manip.baseBody:
                    break

            if manip.baseBody is None:
                manip.baseBody = body

            manip.joints.reverse()

        if not manip.activeBodies:

            body = manip.effectorBody

            while body.parentJoint or body.offsetFrom:
                manip.activeBodies[body.index] = 1
                if body.parentJoint:
                    body = body.parentJoint.body1
                else:
                    body = body.offsetFrom

            manip.activeBodies[body.index] = 1

        manip.index = len(self.manipulators)
        self.manipulators.append(manip)

        if len(manip.joints) == 2 or len(manip.joints) == 3:

            jvalues = numpy.zeros(len(self.joints))
            xforms = self.transforms(jvalues)
            binv = xforms[manip.baseBody.index].inverse()
            fk = binv * self.manipulatorFK(xforms, manip)

            j1 = manip.joints[0]
            j2 = manip.joints[1]

            j1pos = binv * xforms[j1.offsetFrom.index] * j1.anchor
            j2pos = binv * xforms[j2.offsetFrom.index] * j2.anchor

            manip.initJointPos = j1pos

            if len(manip.joints) == 2:
                j3pos = fk.pos
            else:
                j3 = manip.joints[2]
                # j3pos expressed in base frame
                j3pos = binv * xforms[j3.offsetFrom.index] * j3.anchor

                # Memorize the position and angle of the wrist in the
                # frame of the manipulator
                manip.wristPos = fk.transformInv(j3pos)
                manip.a3 = fk.theta
                

            d21 = j2pos - j1pos
            dw2 = j3pos - j2pos

            manip.l1 = gutil.norm2D(d21)
            manip.a1 = gutil.angle2D(d21)

            manip.l2 = gutil.norm2D(dw2)
            manip.a2 = gutil.angle2D(dw2)

        return manip


    def lookupBody(self,name):
        """Return the Body whose name matches the one given."""
        for o in self.bodies:
            if o.name == name:
                return o
        return None

    def lookupJoint(self,name):
        """Return the Joint whose name matches the one given."""
        for o in self.joints:
            if o.name == name:
                return o
        return None

    def lookupManipulator(self,name):
        """Return the Manipulator whose name matches the one given."""
        for o in self.manipulators:
            if o.name == name:
                return o
        return None

    def buildLookups(self):
        """Call this method when you are done adding objects to build
        an index of data structures used by other methods."""
        nbodies = len(self.bodies)
        njoints = len(self.joints)

        bbarray = numpy.zeros((nbodies,nbodies),dtype='bool')
        bjarray = numpy.zeros((nbodies,njoints),dtype='bool')
        
        for i in range(nbodies):
            bi = self.bodies[i]
            for j in range(nbodies):
                bj = self.bodies[j]
                bbarray[i,j] = self.isAncestorOf(bi, bj)
            for j in range(njoints):
                jj = self.joints[j]
                bjarray[i,j] = self.bodyDependsOnJoint(bi, jj)

        self._isAncestorOf = bbarray
        self._bodyDependsOnJoint = bjarray

    def isAncestorOf(self, rootBody, leafBody):
        """Returns true if rootBody can be reached from leafBody by
        following body parents."""
        if self._isAncestorOf is not None:
            return self._isAncestorOf[rootBody.index, leafBody.index]
        if leafBody.offsetFrom is None:
            return False
        elif leafBody.offsetFrom is rootBody:
            return True
        else:
            return self.isAncestorOf(rootBody, leafBody.offsetFrom)

    def bodyDependsOnJoint(self, body, joint):
        """Return true if the position of the given body is determined
        in part by the given joint."""
        if self._bodyDependsOnJoint is not None:
            return self._bodyDependsOnJoint[body.index, joint.index]
        else:
            return joint.body2 is body or self.isAncestorOf(joint.body2, body)

    def transforms(self, jvalues=None, 
                   xform=gutil.Transform2D(), activeBodies=None):
        """Given a list or array of joint angle values (one for each
        joint), and an optional transform for the root elements of the
        model, returns a list of transformations for each Body of the
        model giving their pose in the world frame."""
        assert(jvalues is None or len(jvalues) == len(self.joints))

        # construct rel transforms
        relXforms = []

        for body in self.bodies:
            if not activeBodies or activeBodies.has_key(body.index):
                relXforms.append(body.xform)
            else:
                relXforms.append(gutil.Transform2D())

        if jvalues is not None:
            for i in range(len(jvalues)):
                joint = self.joints[i]
                bi2 = joint.body2.index
                if not activeBodies or activeBodies.has_key(bi2):
                    assert( self.bodies[bi2].offsetFrom is joint.body1 )
                    t = joint.anchor
                    jx = gutil.Transform2D(t) * gutil.Transform2D(jvalues[i]) * gutil.Transform2D(-t)
                    if (joint.offsetFrom is joint.body1):
                        relXforms[bi2] = jx * relXforms[bi2]
                    else:
                        relXforms[bi2] = relXforms[bi2] * jx

        absXforms = []

        for i in range(len(self.bodies)):
            body = self.bodies[i]
            if not activeBodies or activeBodies.has_key(body.index):
                if body.offsetFrom is None:
                    absXforms.append(xform * relXforms[i])
                else:
                    absXforms.append(absXforms[body.offsetFrom.index] * relXforms[i])
            else:
                absXforms.append(gutil.Transform2D())

        return absXforms



    def jointPosJacobian(self, xforms, body, joint, point, pointInWorldFrame):
        """Given a list of body transformations (returned by
        transforms()), a particular body and joint, and a point,
        determine the change in position of the point given change in
        joint angle. This is a single column of a Jacobian matrix for
        e.g. a particular manipulator. The point can be defined in the
        frame of the given body or the world frame."""
        if not self.bodyDependsOnJoint(body, joint):
            return gutil.point2D(0,0)
        else:
            if pointInWorldFrame:
                worldPoint = point
            else:
                worldPoint = xforms[body.index] * point
            worldAnchor = xforms[joint.offsetFrom.index] * joint.anchor
            return gutil.perp2D(worldAnchor - worldPoint)

    def computeCOM(self, xforms):
        """Given a list of body transformations (returned by
        transforms()), return the position of the overall center of
        mass expressed in the world frame."""
        totalMass = 0
        rval = gutil.point2D(0,0)
        for i in range(len(self.bodies)):
            body = self.bodies[i]
            rval += body.mass * ( xforms[i] * body.com )
            totalMass += body.mass

        if totalMass:
            rval /= totalMass
        return rval

    def manipulatorFK(self, xforms, manip):
        """Given a list of body transformations (returned by
        transforms()) and a manipulator, return the transform from the
        manipulator frame to the world frame.  The translation
        component of this transformation is equivalent to the position
        of its origin, expressed in the world frame."""
        return xforms[manip.effectorBody.index] * manip.xform


    def getSupport(self, xforms, bodies):
        """Get the supporting interval by taking the hull of
        projections onto the world x-axis, given a list of body
        transformations (returned by transforms()) and a list of
        bodies."""
        support = gutil.Interval1D()
        axis = gutil.point2D(1,0)
        for b in bodies:
            x = xforms[b.index]
            for s in b.shapes:
                xs = gutil.Shape2D(s.polygon, xform = x*s.xform)
                p = xs.project(axis)
                support = support.hull(p)
        return support

    def manipulatorPosJacobian(self, xforms, manip):
        """Given a list of body transformations (returned by
        transforms()) and a manipulator, return the kinematic Jacobian
        of the position of the manipulator. The matrix returned always
        has two rows, and has as many columns as there are joints
        controlling the manipulator."""
        njoints = len(manip.joints)
        J = numpy.zeros((2,njoints),dtype='f')
        for col in range(njoints):
            J[:,col] = self.jointPosJacobian(xforms, 
                                             manip.effectorBody,
                                             manip.joints[col],
                                             manip.xform.pos,
                                             False)
        return J

    def clampToLimits(self, jvalues):
        """Given a list or array of joint values, modify it so that
        all angles are clamped to the limits of each joint, if they
        are defined."""
        assert(len(jvalues) == len(self.joints))

        for i in range(len(jvalues)):
            jvalues[i] = self.joints[i].limits.clamp(jvalues[i])

    def analyticIK(self, manip, ikmode, desired, jvalues,
                   baseXform=gutil.Transform2D()):

        """Analytic inverse kinematics. Given a manipulator,
        compute the joint angles which attain a desired position. The
        jvalues argument is a list or array of joint angles which is
        used for *both* input and ouput. The ikmode argument can be
        one of the following values:

          - KINMODEL_IK_BPOS: desired is a 2D point, expressed
            relative to the base transformation provided.

          - KINMODEL_IK_BXFORM: desired is a 2D rigid transformation
            expressed relative to the base transformation provided.

          - KINMODEL_IK_WPOS: desired is a 2D point, expressed
            relative to the world frame.

          - KINMODEL_IK_WXFORM: desired is a 2D rigid transformation
            expressed relative to the world frame.
            
          - KINMODEL_IK_IGNORE: causes this function to be a no-op.

        For KINMODEL_IK_BPOS and KINMODEL_IK_WPOS, the manipulator
        provided must be controlled by exactly 2 joints. For
        KINMODEL_IK_BXFORM and KINMODEL_IK_WXFORM, the manipulator
        must be controlled by exactly 3 points.
            
        The function returns True if an exact solution within the
        joint limits could be found, and False otherwise. In both
        cases, the jvalues passed in will be modified to the best
        solution found."""

        if ikmode == KINMODEL_IK_IGNORE:
            return

        assert(len(jvalues) == len(self.joints))

        if ikmode == KINMODEL_IK_BPOS:
            desired = baseXform * desired
            ikmode = KINMODEL_IK_WPOS
        elif ikmode == KINMODEL_IK_BXFORM:
            assert(len(manip.joints) == 3)
            desired = baseXform * desired
            ikmode = KINMODEL_IK_WXFORM

        if ikmode == KINMODEL_IK_WPOS:
            if len(manip.joints) != 2:
                raise Exception('analytic position IK only supported '
                                'for 2-joint manipulators')
        elif ikmode == KINMODEL_IK_WXFORM:
            if len(manip.joints) != 3:
                raise Exception('analytic position/rotation IK only supported '
                                'for 3-joint manipulators')
        else:
            raise Exception('unsupported IK mode in KinModel.analyticIK')

        xforms = self.transforms(jvalues, xform=baseXform,
                                 activeBodies=manip.activeBodies)

        if len(manip.joints) == 2:
            p = xforms[manip.baseBody.index].transformInv(desired)
        else:
            # get the desired in the base frame
            world_from_ee = desired
            world_from_base = xforms[manip.baseBody.index]
            base_from_ee = world_from_base.inverse() * world_from_ee
            p = base_from_ee * manip.wristPos
            
        p -= manip.initJointPos

        dsqr = gutil.normSqr2D(p)
        d = math.sqrt(dsqr)

        l1 = manip.l1
        l2 = manip.l2
        a1 = manip.a1
        a2 = manip.a2

        # try to cope with unreachables
        if d > l1 + l2:
            d = l1+l2
            rval = False
        elif d < math.fabs(l1-l2):
            d = math.fabs(l1-l2)
            rval = False
        else:
            rval = True

        cosbeta = (l1*l1 + l2*l2 - dsqr)/(2*l1*l2) 
        cosgamma = (l1*l1 + dsqr - l2*l2)/(2*l1*d) 

        # need clamp here to deal with roundoff error
        beta = math.acos( gutil.clamp(cosbeta, -1, 1) )
        gamma = math.acos( gutil.clamp(cosgamma, -1, 1) )

        alpha = gutil.angle2D(p)

        best = None
        bestDist = 0
        
        for s in [-1.0, 1.0]:
                
            t1 = alpha - s*gamma - a1
            t2 = math.pi - s*beta + a1 - a2

            if len(manip.joints) == 3:
                a3 = manip.a3
                t3 = base_from_ee.theta - t1 - t2  - a3
                current = (t1,t2,t3)
            else:
                current = (t1,t2)

            dist = 0.0
            for i in range(len(manip.joints)):
                ti = current[i]
                ci = jvalues[manip.joints[i].index]
                dist += math.fabs(gutil.angleDiff(ti,ci))

            if best is None or dist < bestDist:
                best = current
                bestDist = dist

        for i in range(len(manip.joints)):

            # TODO: debug to see if this actually works well with limits
            lim = manip.joints[i].limits

            if lim.empty():
                ctr = 0.0
            else:
                ctr = 0.5*(lim.lower + lim.upper)
            
            lo = ctr - math.pi
            hi = ctr + math.pi
            j = gutil.wrap(best[i], lo, hi)

            if not lim.empty() and not lim.contains(j):
                j = lim.clamp(j)
                rval = False

            jvalues[manip.joints[i].index] = j

        return rval

    def manipulatorIK(self, manip, ikmode, desired, jvalues, 
                      baseXform=gutil.Transform2D(), maxIter=100):


        """Jacobian-based inverse kinematics. Given a manipulator,
        compute the joint angles which attain a desired position. The
        jvalues argument is a list or array of joint angles which is
        used for *both* input and ouput. The ikmode argument can be
        one of the following values:

          - KINMODEL_IK_BPOS: desired is a 2D point, expressed
            relative to the base transformation provided.

          - KINMODEL_IK_BXFORM: desired is a 2D rigid transformation
            expressed relative to the base transformation provided.

          - KINMODEL_IK_WPOS: desired is a 2D point, expressed
            relative to the world frame.

          - KINMODEL_IK_WXFORM: desired is a 2D rigid transformation
            expressed relative to the world frame.
            
          - KINMODEL_IK_IGNORE: causes this function to be a no-op.
            
        This function will not necessarily converge, and may violate
        joint limits. However, it should *not* ever result in an
        *increase* in distance to the desired position or
        transformation, it can be safely called without worrying about
        overwriting joint angles.

        As an optimization, this function calls self.analyticIK() when
        the number of degrees of freedom of the manipulator agrees
        with the degrees of freedom in the desired pose."""

        if ikmode == KINMODEL_IK_IGNORE:
            return
        elif ( ( len(manip.joints) == 2 and
               ( ikmode == KINMODEL_IK_BPOS or 
                 ikmode == KINMODEL_IK_WPOS) ) or
             ( len(manip.joints) == 3 and
               ( ikmode == KINMODEL_IK_BXFORM or
                 ikmode == KINMODEL_IK_WXFORM) ) ):
            return self.analyticIK(manip, ikmode, desired, jvalues, baseXform)

        assert(len(jvalues) == len(self.joints))

        if ikmode == KINMODEL_IK_BPOS:
            desired = baseXform * desired
            ikmode = KINMODEL_IK_WPOS
        elif ikmode == KINMODEL_IK_BXFORM:
            desired = baseXform * desired
            ikmode = KINMODEL_IK_WXFORM
        
        if ikmode == KINMODEL_IK_WPOS:
            jrows = 2
        elif ikmode == KINMODEL_IK_WXFORM:
            jrows = 3
        else:
            raise Exception('unsupported IK mode in KinModel.computeIK')

        jcols = len(manip.joints)
        J = numpy.zeros((jrows, jcols), dtype='f')

        if jrows == 3:
            J[2,:] = 1.0

        err = numpy.zeros(jrows, dtype='f')
        delta = numpy.zeros(jcols, dtype='f')

        alpha = 1e-2
        lastErr = 0
        initErr = 0
        lastParams = [0.0]*jcols

        for iter in range(maxIter):

            xforms = self.transforms(jvalues, xform=baseXform,
                                     activeBodies=manip.activeBodies)

            fk = self.manipulatorFK(xforms, manip)

            posErr = 0
            rotErr = 0

            if ikmode == KINMODEL_IK_WPOS:
                posErr = desired - fk.pos
            elif ikmode == KINMODEL_IK_WXFORM:
                posErr = desired.pos - fk.pos
                rotErr = gutil.angleDiff(fk.theta, desired.theta)
            else:
                raise Exception('unsupported IK mode in KinModel.computeIK')

            posErrMag = gutil.norm2D(posErr)

            if posErrMag < self.ikPosTol and rotErr < self.ikRotTol:
                return True

            err[0:2] = posErr[:]

            if jrows == 3:
                err[2] = rotErr
            
            curErr = numpy.linalg.norm(err)

            if iter == 0:
                initErr = curErr
                lastErr = curErr
                for j in range(jcols):
                    lastParams[j] = jvalues[manip.joints[j].index]
            elif curErr > lastErr:
                alpha = max(alpha*0.1, 1e-5)
                for j in range(jcols):
                     jvalues[manip.joints[j].index] = lastParams[j]
                curErr = lastErr
                continue
            else:
                alpha = min(alpha*2, 1)
                for j in range(jcols):
                    lastParams[j] = jvalues[manip.joints[j].index]
                lastErr = curErr

            if iter+1 == maxIter:
                break
                
            J[0:2,:] = self.manipulatorPosJacobian(xforms, manip)

            #delta[:] = alpha*numpy.linalg.lstsq(numpy.dot(J.transpose(), J), 
            #                                    numpy.dot(J.transpose(), err))[0]

            delta[:] = alpha*numpy.linalg.lstsq(J, err)[0]

            for j in range(len(manip.joints)):
                joint = manip.joints[j]
                jvalues[joint.index] += delta[j]
                

        xforms = self.transforms(jvalues, xform=baseXform,
                                 activeBodies=manip.activeBodies)

        return False
                    


    def forEachShape(self, xforms, func):
        """Given a list of body transformations (returned by
        transforms()) and a callback function func(body, shape), calls
        the callback once for each shape in each body, transformed
        with the appropriate transformation."""
        assert(len(xforms) == len(self.bodies))
        allShapes = []

        for i in range(len(self.bodies)):
            body = self.bodies[i]
            for shape in body.shapes:
                func(body, gutil.Shape2D(shape.polygon,
                                         name=shape.name,
                                         xform=xforms[i]*shape.xform,
                                         color=shape.color))

    def drawShape(self, index, shape, filled=True, colored=True, overrideColors={}):
        """Helper method used by self.draw()."""
        if colored and overrideColors and overrideColors.has_key(index):
            gl.glColor(overrideColors[index])
            colored=False
        shape.draw(colored=colored, filled=filled)

    def transformedShapes(self, xforms):
        """Return a list of all shapes, transformed by the given
        transformations."""
        rval = []
        self.forEachShape(xforms, lambda b, s: rval.append(s))
        return rval

    def transformedBodyShapes(self, xforms):
        """Return a list of (body,shape) pairs, with shapes
        transformed by the given transformations."""
        rval = []
        self.forEachShape(xforms, lambda b, s: rval.append((b,s)))
        return rval
    
    def draw(self, xforms, filled=True, colored=True, overrideColors={}):
        """Draw this model."""
        self.forEachShape(xforms, 
                          lambda b, s: self.drawShape(b.index, s, 
                                                      filled, colored, 
                                                      overrideColors))


    def drawSkeleton(self, xforms, drawBody, drawJoint=None, drawManip=None):
        """Draw a skeleton for this model. drawBody, drawJoint, and
        drawManip are callbacks which are called with the position of
        body origins, joints, and manipulators."""

        if drawJoint is None:
            drawJoint = drawBody

        if drawManip is None:
            drawManip = drawBody
    
        gl.glColor3f(0.0, 0.0, 0.0)

        gl.glBegin(gl.GL_LINES)
        for i in range(len(self.bodies)):
            body = self.bodies[i]
            for j in body.childJoints:
                bi = j.offsetFrom.index
                gl.glVertex2fv(xforms[i].pos)
                gl.glVertex2fv(xforms[bi] * j.anchor)
            if body.parentJoint:
                bi = body.parentJoint.offsetFrom.index
                gl.glVertex2fv(xforms[i].pos)
                gl.glVertex2fv(xforms[bi] * body.parentJoint.anchor)
        for m in self.manipulators:
            t = self.manipulatorFK(xforms, m)
            bi = m.effectorBody.index
            gl.glVertex2fv(xforms[bi].pos)
            gl.glVertex2fv(t.pos)
                
        gl.glEnd()

        for i in range(len(self.bodies)):
            body = self.bodies[i]
            drawBody(xforms[i].pos)

        for j in self.joints:
            bi = j.offsetFrom.index
            drawJoint(xforms[bi] * j.anchor)

        for m in self.manipulators:
            t = self.manipulatorFK(xforms, m)
            drawManip(t.pos)
            


