RRT_STATUS_FAIL_MAX_ITER  = -2
RRT_STATUS_FAIL_MAX_NODES = -1
RRT_STATUS_SUCCESS        =  0
RRT_STATUS_RUNNING        =  1

RRT_EXTEND_TRAPPED       = -1
RRT_EXTEND_ADVANCED      = 0
RRT_EXTEND_REACHED_STATE = 1
RRT_EXTEND_REACHED_GOAL  = 2

RRT_TREE_INIT = 0
RRT_TREE_GOAL = 1


"""Module to implement a base class for RRT planners. Written for E91
Humanoid Robotics, Spring 2013, by Matt Zucker."""

def rrtStatusString(value):
    """Converts an RRT object's self.status to a string."""
    if value < -2 or value > 2:
        return '[INVALID STATUS]'
    else:
        return ['RRT_STATUS_FAIL_MAX_ITER',
                'RRT_STATUS_FAIL_MAX_NODES',
                'RRT_STATUS_SUCCESS',
                'RRT_STATUS_RUNNING'][value+2]

######################################################################

class RRTNode:

    """Class to encapsulate an RRT node. It has a state (self.state,
    any type of object), a parent RRTNode or None (self.parent), as
    well as unique index (self.index) within the tree."""

    def __init__(self, state, parent=None, index=-1, tree=RRT_TREE_INIT):
        """Initialize this RRTNode."""
        self.state = state
        self.parent = parent
        self.index = index
        self.tree = tree

######################################################################

class RRT:

    """Abstract base class to implement much of the functionality of
    RRT search."""

    def __init__(self, maxNodes=0, maxIter=0):
        """Initialize this object. If maxNodes and maxIter are not
        specified beyond the default value of zero, they will be
        ignored, and the RRT can potentially run forever, so it is
        wise to specify non-zero values for these."""
        self.maxNodes = maxNodes
        self.maxIter = maxIter
        self.useConnectHeuristic = True
        self.bidirectional = False
        self.initialize()
        
    def _insertNode(self, tree, node):
        """Helper function to insert a node into the tree. You should
        call insertState from your subclass's self.extend method
        instead of calling this explicitly."""
        self.tree = tree
        idx = len(self.nodes[self.tree])
        node.index = idx
        node.tree = tree
        self.nodes[tree].append(node)
        self.nearestNode[self.tree] = node
        self.nodeAdded(tree, node)
        return node

    def insertState(self, state):
        """Call this function (generally zero or one times) inside of
        your subclass's self.extend method to add a state to the
        tree. The current tree and parent is deduced automatically
        from the nearest neighbor or previous extensions."""
        self._insertNode(self.tree, 
                         RRTNode(state, self.nearestNode[self.tree]))

    def nodeAdded(self, tree, node):
        """Callback called whenever a node is added to the tree. This
        function is a no-op in the base class, but can be redefined in
        subclasses. The typical reason for doing so would be to add
        states to an index independently managed by the subclass for
        faster nearest neighbor search."""
        pass 

    def _unimplemented(self, method):
        """Raise an exception if a subclass has not redefined a key
        method of this class: randomState, findNearestIndex, or
        extend."""
        raise Exception('RRT.'+method+' is unimplemented!')
        
    def randomState(self):
        """Redefine this method in your subclass to return a randomly
        sampled state for the given tree."""
        self._unimplemented('randomState')

    def findNearestIndex(self, state, tree, nodes):
        """Redefine this method in your subclass to return the index
        of the node whose state is closest to the one given. Note that
        linear search over the nodes array may be slow in Python, so
        you might want to consider managing your own index in your
        subclass of RRT."""
        self._unimplemented('findNearestIndex')

    def extend(self, tree, origState, sampledState):
        """Redefine this method in your subclass to attempt to add a
        state extending from origState (the nearest state in the tree)
        to the randomly sampled sampledState. Your method should
        return one of the following values:

          - RRT_EXTEND_TRAPPED: no states could be added to the tree
            due to obstacles or otherwise invalid edges.

          - RRT_EXTEND_REACHED_GOAL: the extension was successful, and
            the new state is a goal state for the search, so the
            overall search is successful as well.

          - RRT_EXTEND_REACHED_STATE: the extension succeeded, and the
            new state is equal to the sampledState that was passed in.

          - RRT_EXTEND_ADVANCED: the extension succeded, but the new
            state is neither a goal state nor the sampledState that
            was passed in.

        If you want add a state to the tree, do so by calling
        self.insertState() in this method."""
        self._unimplemented('extend')

    def initialize(self, initState=None, goalState=None):
        """Initialize the search. This clears the tree and inserts the
        given state into the root node of the new tree."""

        self.nodes = [[],[]]
        self.solution = []
        self.iter = 0

        if initState is not None:
            self._insertNode(RRT_TREE_INIT, RRTNode(initState))

        if goalState is not None:
            self._insertNode(RRT_TREE_GOAL, RRTNode(goalState))


        if self._isBidirectional():
            self.tree = RRT_TREE_GOAL
        else:
            self.tree = RRT_TREE_INIT

        self.status = RRT_STATUS_RUNNING
        self.merging = False

        self._initStep()

    def _initStep(self):

        if self._isBidirectional() and not self.merging:
            self.tree = 1-self.tree

        self.sample = None
        self.nearestNode = [None, None]
        self.merging = False

    def _growTowardsTree(self):

        self.sample = self.nearestNode[self.tree].state
        self.tree = 1-self.tree
        self.merging = True

    def _bidirectionalFinished(self):

        node = self.nearestNode[RRT_TREE_INIT]

        while node:
            self.solution.append(node.state)
            node = node.parent

        self.solution.reverse()

        node = self.nearestNode[RRT_TREE_GOAL].parent

        while node:
            self.solution.append(node.state)
            node = node.parent

        self.sample = None
        self.nearestNode = [None, None]
        self.merging = False
        self.status = RRT_STATUS_SUCCESS

    def _forwardFinished(self):

        node = self.nearestNode[self.tree]
        
        while node:
            self.solution.append(node.state)
            node = node.parent

        self.solution.reverse()

        self.sample = None
        self.nearestNode = [None, None]
        self.status = RRT_STATUS_SUCCESS

    def _isBidirectional(self):
        return len(self.nodes[RRT_TREE_GOAL])>0

    def totalNodes(self):
        return len(self.nodes[RRT_TREE_INIT]) + len(self.nodes[RRT_TREE_GOAL])

    def step(self):
        """Perform a single step of the RRT algorithm. You can examine
        self.status to see how the search is proceeding. It may take
        on one of many values:

          - RRT_STATUS_RUNNING: the search is running

          - RRT_STATUS_FAIL_MAX_ITER: the maximum number of sampling
            iterations has been reached.

          - RRT_STATUS_FAIL_MAX_NODES: the maximum number of nodes has
            been reached.
        """

        if self.status != RRT_STATUS_RUNNING:

            raise Exception('invalid status in RRT.step!')

        elif self.sample is None:

            if self.maxIter>0 and self.iter>=self.maxIter:
            
                self.status = RRT_STATUS_FAIL_MAX_ITER
                
            else:

                self.iter += 1
                self.sample = self.randomState()
                self.nearestNode[self.tree] = None
                
        elif self.nearestNode[self.tree] is None:

            idx = self.findNearestIndex(self.sample,
                                        self.tree,
                                        self.nodes[self.tree])

            self.nearestNode[self.tree] = self.nodes[self.tree][idx]

        else:

            extendStatus = self.extend(self.tree,
                                       self.nearestNode[self.tree].state,
                                       self.sample)

            if extendStatus == RRT_EXTEND_TRAPPED:

                self._initStep()

            elif extendStatus == RRT_EXTEND_ADVANCED:

                if self._isBidirectional() and not self.merging:
                    self._growTowardsTree()
                elif not self.useConnectHeuristic:
                    self._initStep()

            elif extendStatus == RRT_EXTEND_REACHED_STATE:

                if self._isBidirectional():
                    if self.merging:
                        self._bidirectionalFinished()
                    else:
                        self._growTowardsTree()
                else:
                    self._initStep()

            elif extendStatus == RRT_EXTEND_REACHED_GOAL:

                if self._isBidirectional():
                    raise Exception('RRT.extend() returned'
                                    'RRT_EXTEND_REACHED_GOAL'
                                    'for bidirectional tree')
                else:
                    self._forwardFinished()

            if ( self.status != RRT_STATUS_SUCCESS and 
                 self.maxNodes > 0 and
                 self.totalNodes() > self.maxNodes ):

                self.status = RRT_STATUS_FAIL_MAX_NODES

    def search(self, initState=None, goalState=None, maxNodes=0, maxIter=0):
        """Calls self.step() until the status is successful or
        failed."""

        if initState is not None:
            self.initialize(initState, goalState)

        if maxNodes > 0:
            self.maxNodes = maxNodes

        if maxIter > 0:
            self.maxIter = maxIter

        while self.status > 0:
            self.step()


"""

Ta.init(q_init)
Tb.init(q_goal)

for k = 1 to K do
  q_rand = RANDOM_CONFIG()
  if not (EXTEND(Ta,q_rand) == TRAPPED) then
    if (CONNECT(Tb,q_new) == REACHED) then
      return PATH(Ta,Tb)
  SWAP(Ta, Tb)

Return FAILURE

"""
