# Collision checking demo for E91 - Humanoid Robotics
# Author: Matt Zucker

# Assume that some necessary E91 Python modules live in the sibling
# directory "common", and tell python where to reach them.
import sys
sys.path.append('../common')

# Pull in a bunch stuff from other modules.
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

from GeomUtil import *
from GlutApp import *

import math
import numpy

# Subclass of GlutApp to display our collision checking demo.
class CollisionApp(GlutApp):

    def __init__(self):

        # Initialize the parent class
        GlutApp.__init__(self, title="Collision Test")

        # Set up a bounding box for the window in world coords.
        self.bounds = Box2D(-10,-10,10,10)

        # Make some lines and circles in the world.
        self.lines = [ ( point2D(-5, 0), point2D(5,  0) ), 
                       ( point2D(-5, 5), point2D(5, -5) ) ]

        self.circles = [ (point2D(0,  5), 1),
                         (point2D(0, -5), 1) ]

        # Create lists of random shapes, along with velocities to move
        # them by at each timestep. The DragManager will be
        # responsible for allowing the user to drag the shapes around.

        self.nshapes = 12
        self.shapes = []
        self.velocities = []
        self.dragManager = DragManager()

        for i in range(self.nshapes):

            nsides = numpy.random.randint(3, 7)
            dtheta = 2*math.pi / nsides

            angles = ( numpy.arange(nsides)*dtheta + 
                       numpy.random.rand(nsides)*0.75*dtheta )
            angles = numpy.fmod(angles, 2*math.pi)
            angles.sort()

            r = 1 + numpy.random.rand()

            points = []

            for theta in angles:
                points.append(point2DFromPolar(r, theta))

            color = numpy.random.rand(3)*0.5 + 0.5
            color[numpy.random.randint(3)] = 0

            xform = Transform2D(self.bounds.random())

            self.velocities.append((randpm(0.05), randpm(0.05), randpm(0.005)))

            shape = Shape2D(points, xform=xform, color=color)

            self.shapes.append(shape)

            self.dragManager.add(shape)

        # Initial collision check to make sure things draw correctly
        # the first time.
        self.checkCollisions()

        # We are initially animating
        self.animating = True

        # We do not want to draw bounding circles for shapes by default
        self.drawBoundingCircles = False

        # Set a timer callback to check collisions and move objects,
        # at a 5m time interval.
        self.delay = 5
        self.setTimer(self.delay)


    # Timer callback called at 5ms interval.
    def timer(self, value):

        # If animating and not dragging, advance the shapes by their
        # velocities.
        if self.animating and not self.dragManager.isDragging():

            for i in range(self.nshapes):

                shape = self.shapes[i]
                vel = self.velocities[i]
                (x,y) = shape.xform.pos
                shape.xform.pos += vel[0:2]
                shape.xform.pos = self.bounds.wrap( shape.xform.pos )
                shape.xform.theta += vel[2]

        # Perform a collision check.
        self.checkCollisions()

        # Re-register the timer to fire again after 5ms
        self.setTimer(self.delay)
        
        # Tell the display to update.
        glutPostRedisplay()


    # Callback to handle keyboard input.
    def keyboard(self, key, x, y):

        # Convert to lowercase
        key = key.lower()

        if key == ' ':
            # Space toggles animation
            self.animating = not self.animating
        elif key == 'c':
            # 'C' toggles drawing bounding circles
            self.drawBoundingCircles = not self.drawBoundingCircles
        else:
            # Let the parent class handle anything else.
            GlutApp.keyboard(self, key, x, y)
                

    # Check each shape to see whether it is colliding with another
    # shape or a static circle/line.
    def checkCollisions(self):

        # Initialize collisions to false for all shapes.
        self.hits = [False]*self.nshapes

        # For each shape: 
        for i in range(self.nshapes):

            shape = self.shapes[i]

            # Check against all lines:
            for l in self.lines:
                if ( not self.hits[i] and 
                     shape.overlapsLine(l[0], l[1]) ):
                    self.hits[i] = True

            # Check against all circles:
            for c in self.circles:
                if ( not self.hits[i] and
                     shape.contains(c[0], c[1]) ):
                    self.hits[i] = True

            # Check against all other shapes (with index less than
            # this one's so we don't test each pair twice):
            for j in range(i):
                if ( (not self.hits[i] or not self.hits[j])  and
                     shape.overlaps(self.shapes[j]) ):
                    self.hits[i] = True
                    self.hits[j] = True

    # Called when window is resized to setup OpenGL projection matrix.
    def reshape(self, w, h):
        self.setupOrtho2D(w, h, self.bounds)

    # Called when mouse is pressed or released -- just deals with the
    # DragManager.
    def mouse(self, button, state, x, y):
        if button == GLUT_LEFT:
            point = self.unproject2D(x,y)
            if state == GLUT_DOWN:
                self.dragManager.mouseDown(point)
            elif state == GLUT_UP:
                self.dragManager.mouseUp(point)

    # Called when mouse is moved while button is clicked -- just deals
    # with the DragManager.
    def motion(self, x, y):
        if self.dragManager.isDragging():
            point = self.unproject2D(x,y)
            self.dragManager.mouseMoved(point)

    # Callback to actually do the drawing with OpenGL calls.
    def display(self):
        
        # Clear the screen
        glClear(GL_COLOR_BUFFER_BIT)

        # Make sure we are dealing with the modelview matrix and not
        # the projection matrix.
        glMatrixMode(GL_MODELVIEW)

        # Draw lines:
        glColor3ub(127,127,127)
        glBegin(GL_LINES)
        for l in self.lines:
            glVertex2fv(l[0])
            glVertex2fv(l[1])
        glEnd()

        # Draw circles:
        for c in self.circles:
            self.drawCircle(c[0], c[1])

        # Draw shapes:
        for i in range(self.nshapes):
            self.shapes[i].draw(filled = not self.hits[i])

        # Draw bounding circles/centroids
        if self.drawBoundingCircles:

            for shape in self.shapes:
                glColor(0.5*shape.color)
                self.drawCircle(shape.xform * shape.polygon.centroid,
                                shape.polygon.radius, filled=False)

            glBegin(GL_POINTS)
            for shape in self.shapes:
                glColor(0.5*shape.color)
                glVertex2fv(shape.xform * shape.polygon.centroid)
            glEnd()

        # Swap the draw buffers to actually display the image.
        glutSwapBuffers()

# Very short main function just prints a message and runs the app.
if __name__ == '__main__':
    print "Space toggles animation, 'C' toggles circles."
    print "Click & drag to move shapes."
    a = CollisionApp()
    a.run()


