# Dancing man kinematics demo for E91 - Humanoid Robotics
# Author: Matt Zucker

# Assume that some necessary E91 Python modules live in the sibling
# directory "common", and tell python where to reach them.
import sys
sys.path.append('../common')

# Pull in a bunch stuff from other modules.
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

from GeomUtil import *
from GlutApp import *

from KinModel import *

import math
import numpy

# Subclass of GlutApp to display our demo.
class KinDemo(GlutApp):

    def __init__(self):

        # Initialize the parent class
        GlutApp.__init__(self, title="Kinematics test")

        # Set up a bounding box for the window in world coords.
        self.bounds = Box2D(-10,-10,10,10)

        # Now initialize our kinematic model.
        self.model = KinModel('dancingman')
        m = self.model

        # First, the bodies:

        torso = m.add(Body('torso', shapes=[Shape2D(Box2D(-2,-2,2,2))]))

        ul_arm = m.add(Body('ul_arm', offsetFrom=torso, 
                            xform=Transform2D(-3.5,1.5),
                            shapes=[Shape2D(Box2D(-1,-0.5,1,0.5))]))
        
        ll_arm = m.add(Body('ll_arm', offsetFrom=ul_arm, 
                            xform=Transform2D(-2.5,0),
                            shapes=[Shape2D(Box2D(-1,-0.5,1,0.5))]))

        l_hand = m.add(Body('l_hand', offsetFrom=ll_arm, 
                            xform=Transform2D(-2.0,0),
                            shapes=[Shape2D(Box2D(-0.5,-0.25,0.5,0.25))]))

        ur_arm = m.add(Body('ur_arm', offsetFrom=torso, 
                            xform=Transform2D(3.5,1.5),
                            shapes=[Shape2D(Box2D(-1,-0.5,1,0.5))]))

        lr_arm = m.add(Body('lr_arm', offsetFrom=ur_arm, 
                            xform=Transform2D(2.5,0),
                            shapes=[Shape2D(Box2D(-1,-0.5,1,0.5))]))

        r_hand = m.add(Body('r_hand', offsetFrom=lr_arm, 
                            xform=Transform2D(2.0,0),
                            shapes=[Shape2D(Box2D(-0.5, -0.25, 0.5, 0.25))]))

        ul_leg = m.add(Body('ul_leg', offsetFrom=torso, 
                            xform=Transform2D(-1.0,-4),
                            shapes=[Shape2D(Box2D(-0.5,-1.5,0.5,1.5))]))

        ll_leg = m.add(Body('ll_leg', offsetFrom=ul_leg, 
                            xform=Transform2D(0,-3.5),
                            shapes=[Shape2D(Box2D(-0.5,-1.5,0.5,1.5))]))

        l_foot = m.add(Body('l_foot', offsetFrom=ll_leg, 
                            xform=Transform2D(0,-2.25),
                            shapes=[Shape2D(Box2D(-0.75,-0.25,0.75,0.25))]))

        ur_leg = m.add(Body('ur_leg', offsetFrom=torso, 
                            xform=Transform2D(1.0,-4),
                            shapes=[Shape2D(Box2D(-0.5,-1.5,0.5,1.5))]))

        lr_leg = m.add(Body('lr_leg', offsetFrom=ur_leg, 
                            xform=Transform2D(0,-3.5),
                            shapes=[Shape2D(Box2D(-0.5,-1.5,0.5,1.5))]))

        r_foot = m.add(Body('r_foot', offsetFrom=lr_leg, 
                            xform=Transform2D(0,-2.25),
                            shapes=[Shape2D(Box2D(-0.75,-0.25,0.75,0.25))]))

        head = m.add(Body('head', offsetFrom=torso, 
                          xform=Transform2D(0,3.5),
                          shapes=[Shape2D(Box2D(-1,-1,1,1))]))

        # Then, the joints:

        l_shoulder = m.add(Joint('l_shoulder', torso, ul_arm, torso, 
                                 point2D(-2.25, 1.5)))

        l_elbow = m.add(Joint('l_elbow', ul_arm, ll_arm, ul_arm, 
                              point2D(-1.25, 0)))

        l_wrist = m.add(Joint('l_wrist', ll_arm, l_hand, ll_arm, 
                              point2D(-1.25, 0)))

        r_shoulder = m.add(Joint('r_shoulder', torso, ur_arm, torso, 
                                 point2D(2.25, 1.5)))

        r_elbow = m.add(Joint('r_elbow', ur_arm, lr_arm, ur_arm, 
                              point2D(1.25, 0)))

        r_wrist = m.add(Joint('r_wrist', lr_arm, r_hand, lr_arm, 
                              point2D(1.25, 0)))

        l_hip = m.add(Joint('l_hip', torso, ul_leg, torso, 
                            point2D(-1, -2.25)))

        l_knee = m.add(Joint('l_knee', ul_leg, ll_leg, ul_leg, 
                             point2D(0, -1.75)))

        l_ankle = m.add(Joint('l_ankle', ll_leg, l_foot, ll_leg, 
                              point2D(0, -1.75)))

        r_hip = m.add(Joint('r_hip', torso, ur_leg, torso, 
                            point2D(1, -2.25)))

        r_knee = m.add(Joint('r_knee', ur_leg, lr_leg, ur_leg, 
                             point2D(0, -1.75)))

        r_ankle = m.add(Joint('r_ankle', lr_leg, r_foot, lr_leg, 
                              point2D(0, -1.75)))

        neck = m.add(Joint('neck', torso, head, torso, 
                           point2D(0, 2.25)))

        # Then, the manipulators for hands and feet

        l_mhand = m.add(Manipulator('l_mhand', l_hand, Transform2D(0, 0)))
        r_mhand = m.add(Manipulator('r_mhand', r_hand, Transform2D(0, 0)))

        l_mfoot = m.add(Manipulator('l_mfoot', l_foot, Transform2D(0, 0)))
        r_mfoot = m.add(Manipulator('r_mfoot', r_foot, Transform2D(0, 0)))

        # Finally, build the index for the model.

        self.model.buildLookups()

        # The time variable is incremented when animating.
        self.time = 0

        # Set up a periodic 5ms timer to animate the man.
        self.delay = 5
        self.setTimer(self.delay, 0)
        
        # Set the base transform for the torso x,y,theta.
        self.baseXform = Transform2D(0, 1)

        # Initialize some initial joint values, one for each joint
        # defined above, and in the same order.
        self.jvalues = numpy.array( [0.5, -1.0, 0.5,
                                     -0.5, 1.0, -0.5,
                                     -0.6, 1.2, -0.6, 
                                     0.6, -1.2, 0.6, 
                                     0.0] )

        # Get the rigid transforms for the model.
        self.xforms = self.model.transforms(self.jvalues, xform=self.baseXform)
        
        
        # Should we do IK for the model?
        self.doIK = True

        # Are IK targets positions (False) or transforms (True)?
        self.ikXform = True
        
        # Should we animate the model dancing?
        self.animating = True

        # Create a DragManager to handle the IK targets, and create a
        # Shape2D corresponding to each one.

        self.dragManager = DragManager()
        for m in self.model.manipulators:
            self.dragManager.add( Shape2D(Box2D(-0.5,-0.5,0.5,0.5), 
                                          color=(0.0, 0, 0.5)) )

        # Compute the forward kinematics to place the draggable shapes
        # in the correct place.
        self.computeFK()

    # Compute the forward kinematics to place the draggable shapes
    # under the end effectors. Assumes that self.xforms holds the
    # correct transformations for the current joint values.
    def computeFK(self):
        for i in range(len(self.model.manipulators)):
            manip = self.model.manipulators[i]
            draggable = self.dragManager.draggables[i]
            xform = self.model.manipulatorFK(self.xforms, manip)
            draggable.shape.xform.pos = xform.pos

    # Compute the inverse kinematics to place the end effectors on top
    # of the draggable shapes, thus modifying self.jvalues. Assumes
    # that self.xforms holds the correct transformations for the
    # current joint values.
    def computeIK(self):

        for i in range(len(self.model.manipulators)):

            manip = self.model.manipulators[i]

            if self.ikXform:
                dpos = self.dragManager.draggables[i].shape.xform
                ikmode = KINMODEL_IK_WXFORM
            else:
                dpos = self.dragManager.draggables[i].shape.xform.pos
                ikmode = KINMODEL_IK_WPOS


            self.model.manipulatorIK( manip, ikmode, dpos, 
                                      self.jvalues, baseXform=self.baseXform, 
                                      maxIter=10 )
        
    # Periodic timer callback.
    def timer(self, value):

        # Make the man dance the sinusoid dance if we are animating.
        if self.animating:
            self.time += 0.1
            self.baseXform.pos[1] = 1 + 0.5*math.sin(self.time)
            self.baseXform.theta = 0.1 * math.sin(2*self.time)
            self.jvalues[-1] = -0.1 * math.sin(2*self.time)

        # Compute IK if necessary.
        if self.doIK:
            self.computeIK()

        # Since IK modified joint values, need to update transforms
        # with new values.
        self.xforms = self.model.transforms(self.jvalues, self.baseXform)

        # Register the timer to call back again.
        self.setTimer(self.delay)

        # Tell GLUT display needs to be redrawn.
        glutPostRedisplay()


    # Called when mouse is pressed or released -- just deals with the
    # DragManager.
    def mouse(self, button, state, x, y):
        if button == GLUT_LEFT:
            point = self.unproject2D(x,y)
            if state == GLUT_DOWN:
                self.dragManager.mouseDown(point)
            elif state == GLUT_UP:
                self.dragManager.mouseUp(point)

    # Called when mouse is moved while button is clicked -- just deals
    # with the DragManager.
    def motion(self, x, y):
        if self.dragManager.isDragging():
            point = self.unproject2D(x,y)
            self.dragManager.mouseMoved(point)

    # Called when window is resized to setup OpenGL projection matrix.
    def reshape(self, w, h):
        self.setupOrtho2D(w, h, self.bounds)

    # Callback to actually do the drawing with OpenGL calls.
    def display(self):

        # Clear the screen
        glClear(GL_COLOR_BUFFER_BIT)

        # Make sure we are dealing with the modelview matrix and not
        # the projection matrix.
        glMatrixMode(GL_MODELVIEW)

        # Draw IK targets
        for d in self.dragManager.draggables:
            d.shape.draw()

        # Draw the model, filled.
        self.model.draw(self.xforms)

        # Draw the skeleton for the model.
        self.model.drawSkeleton(self.xforms, 
                                lambda pos: self.drawCircle(pos, 0.2))

        # Swap the draw buffers to actually display the image.
        glutSwapBuffers()

    # Callback to handle keyboard input.
    def keyboard(self, key, x, y):

        # Convert to lowercase
        key = key.lower()

        if key == 'i':
            # 'I' toggles IK
            self.doIK = not self.doIK
        elif key == ' ':
            # Space toggles animation
            self.animating = not self.animating
        elif key == 'm':
            # 'M'' toggles IK mode.
            self.ikXform = not self.ikXform
        else:
            # Let the parent class handle anything else.
            GlutApp.keyboard(self, key, x, y)

# Very short main function just prints a message and runs the app.
if __name__ == '__main__':
    print "Space toggles dancing, 'I' toggles IK, 'M' toggles IK mode."
    print "Drag the blue squares to move IK targets."
    a = KinDemo()
    a.run()
        
