# 2D RRT demo for E91 - Humanoid Robotics
# Author: Matt Zucker

# Assume that some necessary E91 Python modules live in the sibling
# directory "common", and tell python where to reach them.
import sys
sys.path.append('../common')

# Pull in a bunch stuff from other modules.
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

from GeomUtil import *
from GlutApp import GlutApp

from RRT import *

import math
import numpy

# Subclass of RRT to solve a planning problem for a translating point
# in a maze. The state of the "robot" is simply (x,y).
#
# This particular subclass of RRT uses a big Nx2 numpy array to store
# all states in the tree, and uses numpy to do vectorized operations
# to find the nearest neighbor. It should be slightly faster than
# naive linear search over all the nodes in the tree.
class RRT2D(RRT):

    """Here is our "maze":
     +-------------------------------------------+ 18
     |                     J                     |
     +---+-----------------------------------+---+ 16 
     |   |                                   |   |
     |   |   +---------------+   +---+---+   |   | 14
     |   |   |       A       |   | B |   |   |   |
     |   |   +---------------+   +---+   |   |   | 12 
     |   |                    goal * | C |   |   |
     |   |   +-------+   +---+   +---+   |   |   | 10
     | K |   |   D   |   |   |   | F |   |   | L |
     |   |   +---+---+   +   +   +---+---+   |   | 08 
     |   |   |   |       | E |               |   |
     |   |   + G +   +---+   +   +-------+   |   | 06
     |   |   |   |   | H |   |   |   I   |   |   |
     |   |   +---+   +---+---+   +-------+   |   | 04
     |   | * init                            |   |
     +---+-----------------------------------+---+ 02
     |                     M                     |
     +-------------------------------------------+ 00
     0   0   0   0   0   1   1   1   1   1   2   2
     0   2   4   6   8   0   2   4   6   8   0   2
     """

    def __init__(self, maxNodes=500):

        # Initialize the parent class
        RRT.__init__(self, maxNodes)

        # Set up a bounding box for the maze in world coords.
        self.bounds = Box2D(0, 0, 22, 18)

        # Set up our boxes as Shape2D instances:
        boxdims = [
            ( 4, 12, 12, 14), # A 0
            (14, 12, 16, 14), # B 1 
            (16,  8, 18, 14), # C 2 
            ( 4,  8,  8, 10), # D 3 
            (10,  4, 12, 10), # E 4
            (14,  8, 16, 10), # F 5 
            ( 4,  4,  6,  8), # G 6
            ( 8,  4, 10,  6), # H 7
            (14,  4, 18,  6), # I 8
            ( 0, 16, 22, 18), # J 9
            ( 0,  2,  2, 16), # K 10
            (20,  2, 22, 16), # L 11
            ( 0,  0, 22,  2), # M 12
            ]

        #boxdims = boxdims[-4:]
        
        c = numpy.array([0.7, 0.7, 1.0])
        self.maze = [ Shape2D(Box2D(*x), color=c) for x in boxdims ]

        # Place the inital and goal positions.
        self.initPos = point2D(3,3)
        self.goalPos = point2D(15, 11)
        self.goalRadius = 0.5

        # Disable connect heuristic for this simple demo.
        self.useConnectHeuristic = True

        # Don't be bidirectional at the start
        self.bidirectional = True

        # Probability of sampling goal state
        self.goalProb = 0.1
        
        # Maximum distance to extend tree
        self.maxEdgeDist = 0.5
        
        # Collision checking tolerance for line segments.
        self.collisionTol = 0.15

        # Reset the search
        self.reset()

    # Initialize the RRT and reset the numpy array we use to do
    # fast(er) computation of distances than linear search.
    def reset(self):
        self.points = numpy.zeros((2,self.maxNodes,2))
        if self.bidirectional:
            self.initialize(self.initPos, self.goalPos)
        else:
            self.initialize(self.initPos)

    # Called when a node is added to the tree.  We use this function
    # to insert the new (x,y) state into the numpy array used to
    # search nearest neighbors.
    def nodeAdded(self, tree, node):
        self.points[tree, node.index, :] = node.state

    # Return a random state in the world.
    def randomState(self):
        p = numpy.random.rand()
        if p <= self.goalProb and not self.bidirectional:
            return self.goalPos
        else:
            return self.bounds.random()

    # Use numpy vectorized operations to assemble the list of squared
    # distances from each point in the tree to the current point, and
    # find the minimum. Should be somewhat faster than naive linear
    # search over all of the nodes.
    def findNearestIndex(self, state, tree, nodes):
        activePoints = self.points[tree,0:len(nodes),:]
        distSqr = ( (activePoints[:,0]-state[0])**2 + 
                    (activePoints[:,1]-state[1])**2 )
        idx = distSqr.argmin()

        return idx

    # Check to see if the line segment from point a to point b is free
    # by iterating over all the shapes in the maze.
    def collisionCheck(self, a, b):
        for box in self.maze:
            if box.overlapsLine(a, b, self.collisionTol):
                return False
        return True

    # Compute the state near a which extends towards b, or return b if
    # it is closer than the maximum edge distance. Also returns a
    # boolean indicating whether b was reached or not.
    def computeStateNear(self, a, b):

        d = dist2D(a, b)

        if d < self.maxEdgeDist:
            return b[:], True
        else:
            return interp2D(a, b, self.maxEdgeDist/d), False

    # Extend the tree from the closest state in the tree to the state
    # that was sampled.
    def extend(self, tree, closestState, sampledState):

        # Get the state which moves towards the sample state.
        newState, reached = self.computeStateNear(closestState, sampledState)

        # If the line segment collides, return failure.
        if not self.collisionCheck(closestState, newState):
            return RRT_EXTEND_TRAPPED

        # Put the new state in the tree.
        self.insertState(newState)

        # Figure out what status to return to the RRT.
        if ( not self.bidirectional and 
             dist2D(newState, self.goalPos) < self.goalRadius ):
            return RRT_EXTEND_REACHED_GOAL
        elif reached:
            return RRT_EXTEND_REACHED_STATE
        else:
            return RRT_EXTEND_ADVANCED

# Subclass of GlutApp to display the RRT.
class RRT2DApp(GlutApp):

    def __init__(self):

        # Initialize the parent class
        GlutApp.__init__(self, title="RRT Test")

        # Construct the rrt instance.
        self.rrt = RRT2D()

        # Set up a timer to animate solving.
        self.delay = 5
        self.animating = False
        self.setTimer(self.delay, 0)

    # Periodic timer to step RRT.
    def timer(self, value):
        if self.animating:
            self.stepRRT()
            if self.rrt.status <= 0:
                self.animating = False
        self.setTimer(self.delay, 0)

    # If still solving, step RRT and ask for redraw.
    def stepRRT(self):
        if self.rrt.status > 0:
            self.rrt.step()
            glutPostRedisplay()

    # Called when window is resized to setup OpenGL projection matrix.
    def reshape(self, w, h):
        self.setupOrtho2D(w, h, self.rrt.bounds)

    # Callback to actually do the drawing with OpenGL calls.
    def display(self):

        # Clear the screen
        glClear(GL_COLOR_BUFFER_BIT)

        # Make sure we are dealing with the modelview matrix and not
        # the projection matrix.
        glMatrixMode(GL_MODELVIEW)

        rrt = self.rrt

        # Draw maze
        for box in rrt.maze:
            box.draw()

        # Draw initial/goal pos
        glColor3f(0.25,1,0.25)
        self.drawCircle(rrt.initPos, rrt.goalRadius)

        glColor3f(1,0.25,0.25)
        self.drawCircle(rrt.goalPos, rrt.goalRadius)

        # Draw sample/nearest
        glColor3f(0.75,0,0.75)

        if rrt.sample is not None:
            self.drawCircle(rrt.sample, rrt.collisionTol)

        if rrt.nearestNode[rrt.tree] is not None:
            self.drawCircle(rrt.nearestNode[rrt.tree].state, 
                            2*rrt.collisionTol)

        # Draw solution:
        if len(rrt.solution):
            glLineWidth(8.0)
            glColor3f(1,0,1)
            glBegin(GL_LINE_STRIP)
            for state in rrt.solution:
                glVertex2fv(state)
            glEnd()
            glLineWidth(2.0)

        glColor3f(0,0,0.5)

        # Draw search tree states:
        for nlist in rrt.nodes:
            for node in nlist:
                self.drawCircle(node.state, rrt.collisionTol)

        # Draw search tree segments:
        glBegin(GL_LINES)
        for nlist in rrt.nodes:
            for node in nlist:
                if node.parent:
                    glVertex2fv(node.state)
                    glVertex2fv(node.parent.state)
        glEnd()    

        # Draw text:
        glColor3f(0.0, 0.0, 0.0)
        glRasterPos(0.1, 0.1)
        string = 'iter={0:<4} nodes={1:<4} connect={2:<5} status={3}'.format(
            rrt.iter, rrt.totalNodes(), str(rrt.useConnectHeuristic), 
            rrtStatusString(rrt.status))
        self.drawBitmapString(string)

        # Swap the draw buffers to actually display the image.
        glutSwapBuffers()

    # Callback to handle keyboard input.
    def keyboard(self, key, x, y):

        key = key.lower()

        if key == 's':
            self.animating = False
            self.stepRRT()
        if key == '\r' or key == '\n':
            if self.rrt.status <= 0:
                self.rrt.reset()
            self.rrt.search()
            glutPostRedisplay()
        if key == 'r':
            self.animating = False
            self.rrt.reset()
            glutPostRedisplay()
        if key == 'c':
            self.rrt.useConnectHeuristic = not self.rrt.useConnectHeuristic
            glutPostRedisplay()
        if key == 'b':
            self.rrt.bidirectional = not self.rrt.bidirectional
            self.rrt.reset()
            glutPostRedisplay()
        elif key == ' ':
            self.animating = not self.animating
            glutPostRedisplay()
        else:
            GlutApp.keyboard(self, key, x, y)
                

# Very short main function just prints a message and runs the app.
if __name__ == '__main__':
    print "Press 'S' to step, 'R' to reset, 'C' and to toggle connect heuristic."
    print "Enter finishes search, and Space animates."
    a = RRT2DApp()
    a.run()


