# Kinematic planar arm RRT demo for E91 - Humanoid Robotics
# Author: Matt Zucker

# Assume that some necessary E91 Python modules live in the sibling
# directory "common", and tell python where to reach them.
import sys
sys.path.append('../common')

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

# Pull in a bunch stuff from other modules.
from GeomUtil import *
from GlutApp import *
from KinModel import *

from RRT import *

import math
import numpy

# Subclass of RRT to solve a planning problem for a planar arm among
# circular obstacles. The state of the robot is a vector of joint
# angles, represented as a numpy array.
#
# This particular subclass of RRT uses a big numpy array to store all
# states in the tree, and uses numpy to do vectorized operations to
# find the nearest neighbor. It should be slightly faster than naive
# linear search over all the nodes in the tree.
# 
# This subclass also precomputes a "goal set" of states inside the
# goal region of the search, and samples from that goal set during
# planning.
class RRTArm(RRT):

    def __init__(self, maxNodes=2000, maxIter=0):

        # Initialize the parent class
        RRT.__init__(self, maxNodes, maxIter)

        # Set up a bounding box for the world.
        self.bounds = Box2D(0, 0, 20, 20)

        # Maximum length of RRT edge (map joint angles through
        # Jacobian of straight arm to convert angles to distances).
        self.maxMoveDist = 1.0

        # How many discrete collision checks for each edge?
        self.cchecksPerEdge = 5

        # Minimum distance to enforce between robot and maze during a
        # collision check. To be conservative, this should be greater
        # than or equal to self.maxMoveDist / self.cchecksPerEdge.
        self.collisionTol = 0.2

        # Set up our circular obstacles.
        self.circles = [
            ((12,14),2), 
            ((15,5),4),
            ]

        # Goal position and radius for the end effector.
        self.goalPos = point2D(13, 10.5)
        self.goalRadius = 0.5

        # Build our arm. See KinModel documentaiton for details.
        m = KinModel('arm')
        self.model = m 

        l = 3
        lc = (0.75,0.75,0.75)
        dc = (0.5,0.5,0.5)
        self.nlinks = 5

        baseRect = Box2D(-1,-1,1,1)
        linkRect = Box2D(-0.4,-0.5*l,0.4,0.5*l)

        color = dc

        prev = m.add(Body('base', 
                           xform=Transform2D(5, 1),
                           shapes=[Shape2D(baseRect, color=color)]))

        lp = 2

        margin = 0.25

        self.jlimit = math.pi/2
        limits = Interval1D(-self.jlimit, self.jlimit)

        for i in range(self.nlinks):

            if color is dc:
                color = lc
            else:
                color = dc

            cur = m.add(Body('link{0}'.format(i),
                             offsetFrom=prev,
                             xform=Transform2D(0, 0.5*(lp+l)+margin),
                             shapes=[Shape2D(linkRect,color=color)]))

            m.add(Joint('joint{0}'.format(i),
                        prev, cur, prev,
                        point2D(0, 0.5*(lp+margin)),
                        limits=limits.copy()))

            prev = cur
            lp = l
        
        self.manip = m.add(Manipulator('manip', cur, 
                                       Transform2D(0, 0.5*(lp+margin))))
        
        m.buildLookups()

        # The initial state is simply all zero angles:
        self.initState = numpy.zeros(self.nlinks)

        # Use the straight arm jacobian to map angles to distances of
        # the end effector. This is a good upper bound on the motion
        # of the robot for any angle.
        xforms = m.transforms(self.initState)
        J = m.manipulatorPosJacobian(xforms, self.manip)

        self.distsPerAngle = numpy.fabs( J[0,:] )

        # Build our goal set:
        print 'Please wait, building goal set. This may take some time.'
        goalStates = []
        self.goalStates = []
        self.goalProb = 1

        self.bidirectional = False


        while len(goalStates) < 20:
            state = self.randomState()
            ok, isGoal = self.collisionGoalCheck(state)
            if ok and isGoal:
                goalStates.append(state)


        self.bidirectional = True

        self.goalStates = goalStates
        self.goalProb = 0.15

        # Reset the RRT search.
        self.reset()

    # Initialize the RRT and reset the numpy array we use to do
    # fast(er) computation of distances than linear search.
    def reset(self):
        self.stateLookup = numpy.zeros((2,self.maxNodes,self.nlinks))
        if self.bidirectional:
            self.initialize(self.initState, self.goalStates[numpy.random.randint(len(self.goalStates))])
        else:
            self.initialize(self.initState)

    # Return a randomly sampled state.
    def randomState(self):

        # Sample uniformly within joint limits:
        jvalues = randpm(self.jlimit, self.nlinks)

        # See if we are sampling goal:
        if numpy.random.rand() < self.goalProb and not self.bidirectional:
            # If goal set is built, use that, otherwise, do IK to
            # steer random state above towards goal.
            ngoals = len(self.goalStates)
            if ngoals:
                return self.goalStates[numpy.random.randint(ngoals)]
            else:
                self.model.manipulatorIK(self.manip, 
                                         KINMODEL_IK_WPOS,
                                         self.goalPos,
                                         jvalues, maxIter=100)
                self.model.clampToLimits(jvalues)
        
        # Return our state.
        return jvalues

    # Use numpy vectorized operations to assemble the list of squared
    # distances from each state in the tree to the current state, and
    # find the minimum. Should be somewhat faster than naive linear
    # search over all of the nodes.
    def findNearestIndex(self, state, tree, nodes):
        nnodes = len(nodes)
        smat = numpy.tile(state, (nnodes, 1))
        lmat = self.stateLookup[tree, 0:nnodes, :]
        closest = numpy.sum((smat-lmat)**2, axis=1).flatten().argmin()
        return closest
            
    # Called when a node is added to the tree.  We use this function
    # to insert the new state into the numpy array used to search
    # nearest neighbors.
    def nodeAdded(self, tree, node):
        self.stateLookup[tree,node.index,:] = node.state[:]

    # Compute the state near a which extends towards b, or return b if
    # it is closer than the maximum edge distance. Also returns a
    # boolean indicating whether b was reached or not.
    def computeStateNear(self, a, b):

        adiff = numpy.fabs(b-a)
        d = (adiff * self.distsPerAngle).max()

        if d < self.maxMoveDist:
            return b[:], True
        else:
            return self.interp(a, b, self.maxMoveDist/d), False

    # Helper function to linearly interpolate between two states.
    def interp(self, a, b, u):
        return a + u*(b-a)

    # Extend the tree from the closest state in the tree to the state
    # that was sampled.
    def extend(self, tree, closestState, sampledState):

        # Get the state which moves towards the sample state.
        newState, reached = self.computeStateNear(closestState, sampledState)

        # Do a small number of collision checks for the edge we are
        # attempting to add by interpolating from closestState to
        # newState.
        for i in range(self.cchecksPerEdge):
            u = float(i+1)/self.cchecksPerEdge
            checkState = self.interp(closestState, newState, u)
            ok, isGoal = self.collisionGoalCheck(checkState)
            if not ok:
                return RRT_EXTEND_TRAPPED

        # Insert the state into the tree.
        self.insertState(newState)
        
        # Figure out what status to return to the RRT.
        if ( not self.bidirectional and isGoal ):
            return RRT_EXTEND_REACHED_GOAL
        elif reached:
            return RRT_EXTEND_REACHED_STATE
        else:
            return RRT_EXTEND_ADVANCED

    # Check whether the state collides with the obstacles, and whether
    # the end effector is close enough to the target.
    def collisionGoalCheck(self, jvalues):
        xforms = self.model.transforms(jvalues)
        shapes = self.model.transformedShapes(xforms)
        for shape in shapes:
            for circle in self.circles:
                if shape.contains(circle[0], circle[1]+self.collisionTol):
                    return (False, False)
        xform = self.model.manipulatorFK(xforms, self.manip)
        isGoal = ( dist2D(xform.pos, self.goalPos) < self.goalRadius )
        return (True, isGoal)

# Subclass of GlutApp to display the RRT.
class RRTArmApp(GlutApp):

    def __init__(self):

        # Initialize the parent class
        GlutApp.__init__(self, title="RRT Test")

        # Construct the RRT instance.
        self.rrt = RRTArm()

        # Set up a timer to animate solving.
        self.delay = 50
        self.animating = False
        self.solutionIndex = 0
        self.setTimer(self.delay, 0)

        # Set up some infrastructure for dumping solution to image
        # sequence.
        self.grabPhase = 0
        self.grabber = FrameGrabber('frames', 'image{:04d}.png')


    # Periodic timer to step RRT. We cleverly change the delay based
    # on whether the solver is running or not, to get faster animation
    # when solving, and slower animation when displaying the solution.
    def timer(self, value):
        if self.animating:
            self.delay = 1
            self.stepRRT()
            if self.rrt.status <= 0:
                self.animating = False
        else:
            self.delay = 50
        self.setTimer(self.delay, 0)
        glutPostRedisplay()

    # If still solving, step RRT and ask for redraw.
    def stepRRT(self):
        if self.rrt.status > 0:
            self.rrt.step()
            glutPostRedisplay()

    # Reset the RRT, along with some helper variables.
    def resetRRT(self):
        self.rrt.reset()
        self.solutionIndex = 0
        self.grabPhase = 0
        self.grabber.currentIndex = 0

    # Called when window is resized to setup OpenGL projection matrix.
    def reshape(self, w, h):
        self.setupOrtho2D(w, h, self.rrt.bounds)

    # Helper function to draw the arm at a given state.
    def drawModelAt(self, state, colored=True, filled=True):
        self.xforms = self.rrt.model.transforms(state)
        self.rrt.model.draw(self.xforms, colored=colored, filled=filled)

    # Callback to actually do the drawing with OpenGL calls.
    def display(self):

        # Clear the screen
        glClear(GL_COLOR_BUFFER_BIT)

        # Make sure we are dealing with the modelview matrix and not
        # the projection matrix.
        glMatrixMode(GL_MODELVIEW)

        rrt = self.rrt

        # Draw obstacles
        glColor3f(0.7, 0.7, 1.0)
        for c in rrt.circles:
            self.drawCircle(c[0], c[1])

        # Draw target
        glColor3f(1,0.25,0.25)
        self.drawCircle(rrt.goalPos, rrt.goalRadius)

        # Draw current sample
        glColor3f(0.75,0,0.75)
        if rrt.sample is not None:
            self.drawModelAt(rrt.sample, colored=False, filled=False)

        # Draw the robot, either from the solution or the initial
        # state.
        drawState = rrt.initState

        if len(rrt.solution):
            if self.solutionIndex >= len(rrt.solution):
                self.solutionIndex = 0
                self.grabPhase = 2
            elif self.grabPhase == 0:
                self.grabPhase = 1

            drawState = rrt.solution[self.solutionIndex]
            self.solutionIndex += 1

        self.drawModelAt(drawState)
        self.rrt.model.drawSkeleton(self.xforms, 
                                    lambda pos: self.drawCircle(pos, 0.2))

        # Draw text.
        glColor3f(0.0, 0.0, 0.0)
        glRasterPos(0.1, 0.1)
        string = 'iter={0:<4} nodes={1:<4} connect={2:<5} status={3}'.format(
            rrt.iter, rrt.totalNodes(), str(rrt.useConnectHeuristic), 
            rrtStatusString(rrt.status))
        self.drawBitmapString(string)

        # Dump an image if we're recording the solution.
        if self.grabPhase == 1:
            self.grabber.grabFrame(verbose=True)

        # Swap the draw buffers to actually display the image.
        glutSwapBuffers()

    # Callback to handle keyboard input.
    def keyboard(self, key, x, y):
        key = key.lower()
        if key == 's':
            self.animating = False
            self.stepRRT()
        if key == '\r' or key == '\n':
            if self.rrt.status <= 0:
                self.resetRRT()
            self.rrt.search()
            glutPostRedisplay()
        if key == 'r':
            self.animating = False
            self.resetRRT()
            glutPostRedisplay()
        if key == 'c':
            self.rrt.useConnectHeuristic = not self.rrt.useConnectHeuristic
            glutPostRedisplay()
        elif key == ' ':
            self.animating = not self.animating
            glutPostRedisplay()
        else:
            GlutApp.keyboard(self, key, x, y)
                
# Very short main function just prints a message and runs the app.
if __name__ == '__main__':
    print "Press 'S' to step, 'R' to reset, 'C' and to toggle connect heuristic."
    print "Enter finishes search, and Space animates."
    a = RRTArmApp()
    a.run()


