# Translating/rotating shape RRT demo for E91 - Humanoid Robotics
# Author: Matt Zucker

# Assume that some necessary E91 Python modules live in the sibling
# directory "common", and tell python where to reach them.
import sys
sys.path.append('../common')

# Pull in a bunch stuff from other modules.
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *

from GeomUtil import *
from GlutApp import *

from RRT import *

import math
import numpy

# Subclass of RRT to solve a planning problem for a translating and
# rotating bar in a maze. The state of the "robot" is a 2D rigid
# transform: (x, y, theta) represented using an instance of
# GeomUtil.Transform2D. See the documentation for more details.
#
# This particular subclass illustrates naive linear search over the
# tree to find nearest neighbors. It could perhaps be made faster by
# using vectorized operations from numpy.
class RRTTransRot(RRT):

    """Here is our "maze":
     +-------------------------------------------+ 18
     |                     J                     |
     +---+-----------------------------------+---+ 16 
     |   |                                   |   |
     |   |   +---------------+   +---+---+   |   | 14
     |   |   |       A       |   | B |   |   |   |
     |   |   +---------------+   +---+   |   |   | 12 
     |   |                    goal * | C |   |   |
     |   |   +-------+   +---+   +---+   |   |   | 10
     | K |   |   D   |   |   |   | F |   |   | L |
     |   |   +---+---+   +   +   +---+---+   |   | 08 
     |   |   |   |       | E |               |   |
     |   |   + G +   +---+   +   +-------+   |   | 06
     |   |   |   |   | H |   |   |   I   |   |   |
     |   |   +---+   +---+---+   +-------+   |   | 04
     |   | * init                            |   |
     +---+-----------------------------------+---+ 02
     |                     M                     |
     +-------------------------------------------+ 00
     0   0   0   0   0   1   1   1   1   1   2   2
     0   2   4   6   8   0   2   4   6   8   0   2
     """

    def __init__(self, maxNodes=1000):

        # Initialize the parent class
        RRT.__init__(self, maxNodes)

        # Set up a bounding box for the maze in world coords.
        self.bounds = Box2D(0, 0, 22, 18)

        # Set up our boxes as Shape2D instances:
        boxdims = [
            ( 4, 12, 12, 14), # A 0
            (14, 12, 16, 14), # B 1 
            (16,  8, 18, 14), # C 2 
            ( 4,  8,  8, 10), # D 3 
            (10,  4, 12, 10), # E 4
            (14,  8, 16, 10), # F 5 
            ( 4,  4,  6,  8), # G 6
            ( 8,  4, 10,  6), # H 7
            (14,  4, 18,  6), # I 8
            ( 0, 16, 22, 18), # J 9
            ( 0,  2,  2, 16), # K 10
            (20,  2, 22, 16), # L 11
            ( 0,  0, 22,  2), # M 12
            ]

        c = numpy.array([0.7, 0.7, 1.0])
        self.maze = [ Shape2D(Box2D(*x), color=c) for x in boxdims ]

        # Set up initial state
        self.initState = Transform2D(point2D(4,3))

        # Set up goal position and radius
        self.goalPos = point2D(14, 11)
        self.goalRadius = 0.5

        self.goalState = Transform2D(point2D(self.goalPos))

        self.bidirectional = True

        # Create our "bar robot":
        a = 1.2
        b = 0.2
        bcoords = Box2D(-a,-b,a,b)
        self.robot = Shape2D(bcoords,color=numpy.array([0.5,0.5,0.5]))
        self.robot.xform = self.initState

        # Probability for sampling goal
        self.goalProb = 0.05

        # Angle to weight theta's vs. x-y's in distance metric, see
        # findNearestIndex().
        self.angleWeight = self.robot.polygon.radius

        # Maximum length of RRT edge
        self.maxMoveDist = 0.5
        
        # How many discrete collision checks for each edge?
        self.cchecksPerEdge = 2

        # Minimum distance to enforce between robot and maze during a
        # collision check. To be conservative, this should be greater
        # than or equal to self.maxMoveDist / self.cchecksPerEdge.
        self.collisionTol = 0.25

        # Reset the search.
        self.reset()


    # This simply calls initialize on the RRT with our given initial
    # state.
    def reset(self):
        if self.bidirectional:
            self.initialize(self.initState, self.goalState)
        else:
            self.initialize(self.initState)

    # Return a randomly sampled (x,y,theta) transformation. When
    # sampling from the goal, we just assign a random orientation.
    def randomState(self):
        p = numpy.random.rand()
        if p <= self.goalProb and not self.bidirectional:
            return Transform2D(self.goalPos, randpm(math.pi))
        else:
            return Transform2D(self.bounds.random(), randpm(math.pi))

    # Naive linear search over all nodes in the tree to find the
    # nearest neighbor.
    def findNearestIndex(self, state, tree, nodes):
        bestIndex = None
        bestDist = None
        for i in range(len(nodes)):
            node = nodes[i]
            (pdist, adist) = state.absDistTo(node.state)
            totalDist = pdist + self.angleWeight * adist
            if bestIndex is None or totalDist < bestDist:
                bestIndex = i
                bestDist = totalDist
        return bestIndex

    # Compute the state near a which extends towards b, or return b if
    # it is closer than the maximum edge distance. Also returns a
    # boolean indicating whether b was reached or not.
    def computeStateNear(self, a, b):

        (pdist, adist) = a.absDistTo(b)
        d = pdist + self.robot.polygon.radius * adist

        if d < self.maxMoveDist:
            return b.copy(), True
        else:
            return a.interpTo(b, self.maxMoveDist/d), False


    # Extend the tree from the closest state in the tree to the state
    # that was sampled.
    def extend(self, tree, closestState, sampledState):

        # Get the state which moves towards the sample state.
        newState, reached = self.computeStateNear(closestState, sampledState)

        # Do a small number of collision checks for the edge we are
        # attempting to add by interpolating from closestState to
        # newState.
        for i in range(self.cchecksPerEdge):
            u = float(i+1)/self.cchecksPerEdge
            checkState = closestState.interpTo(newState, u)
            if not self.collisionCheck(newState):
                return RRT_EXTEND_TRAPPED

        # Insert the state into the tree.
        self.insertState(newState)
        
        # Figure out what status to return to the RRT.
        if ( not self.bidirectional and self.isGoalState(newState) ):
            return RRT_EXTEND_REACHED_GOAL
        elif reached:
            return RRT_EXTEND_REACHED_STATE
        else:
            return RRT_EXTEND_ADVANCED

    # Return true if the robot collides with the goal region.
    def isGoalState(self, state):
        self.robot.xform = state
        return self.robot.contains(self.goalPos, self.goalRadius)

    # Compute whether the robot overlaps the maze in the given state.
    def collisionCheck(self, state):

        self.robot.xform = state
        for box in self.maze:
            if self.robot.overlaps(box, self.collisionTol):
                return False
        return True


# Subclass of GlutApp to display the RRT.
class RRTTransRotApp(GlutApp):

    def __init__(self):

        # Initialize the parent class
        GlutApp.__init__(self, title="RRT Test")

        # Construct the rrt instance.
        self.rrt = RRTTransRot()

        # Set up a timer to animate solving.
        self.delay = 50
        self.animating = False
        self.solutionIndex = 0
        self.setTimer(self.delay, 0)

        # Set up some infrastructure for dumping solution to image
        # sequence.
        self.grabPhase = 0
        self.grabber = FrameGrabber('frames', 'image{:04d}.png')

    # Periodic timer to step RRT. We cleverly change the delay based
    # on whether the solver is running or not, to get faster animation
    # when solving, and slower animation when displaying the solution.
    def timer(self, value):
        if self.animating:
            self.delay = 1
            self.stepRRT()
            if self.rrt.status <= 0:
                self.animating = False
        else:
            self.delay = 50
        self.setTimer(self.delay, 0)
        glutPostRedisplay()

    # If still solving, step RRT and ask for redraw.
    def stepRRT(self):
        if self.rrt.status > 0:
            self.rrt.step()
            glutPostRedisplay()

    # Reset the RRT, along with some helper variables.
    def resetRRT(self):
        self.rrt.reset()
        self.solutionIndex = 0
        self.grabPhase = 0
        self.grabber.currentIndex = 0

    # Called when window is resized to setup OpenGL projection matrix.
    def reshape(self, w, h):
        self.setupOrtho2D(w, h, self.rrt.bounds)

    # Callback to actually do the drawing with OpenGL calls.
    def display(self):

        # Clear the screen
        glClear(GL_COLOR_BUFFER_BIT)

        # Make sure we are dealing with the modelview matrix and not
        # the projection matrix.
        glMatrixMode(GL_MODELVIEW)

        rrt = self.rrt

        # Draw maze
        for box in rrt.maze:
            box.draw()

        # Draw initial/goal pos
        glColor3f(0.25,1,0.25)
        self.drawCircle(rrt.initState.pos, rrt.goalRadius)

        glColor3f(1,0.25,0.25)
        self.drawCircle(rrt.goalPos, rrt.goalRadius)

        # Draw sample/nearest
        glColor3f(0.75,0,0.75)
        if rrt.sample:
            self.drawRobotAt(rrt.sample, colored=False, filled=False)
        if rrt.nearestNode[rrt.tree]:
            glLineWidth(6.0)
            self.drawRobotAt(rrt.nearestNode[rrt.tree].state, 
                            colored=False, filled=False)
            glLineWidth(2.0)

        # Draw nodes in tree:
        glColor3f(0.7,0.7,0.7)
        glLineWidth(1.0)

        for nlist in rrt.nodes:
            for node in nlist:
                self.drawRobotAt(node.state, colored=False, filled=False)
        
        glLineWidth(2.0)

        # Draw solution:
        if len(rrt.solution):

            glColor3f(0,0,0.7)
            for state in rrt.solution:
                self.drawRobotAt(state, colored=False, filled=False)
            if self.solutionIndex >= len(rrt.solution):
                self.solutionIndex = 0
                self.grabPhase = 2
            elif self.grabPhase == 0:
                self.grabPhase = 1

            self.drawRobotAt(rrt.solution[self.solutionIndex],
                            colored=True, filled=True)

            self.solutionIndex += 1


        # Draw text:
        glColor3f(0.0, 0.0, 0.0)
        glRasterPos(0.1, 0.1)
        string = 'iter={0:<4} nodes={1:<4} connect={2:<5} status={3}'.format(
            rrt.iter, rrt.totalNodes(), str(rrt.useConnectHeuristic), 
            rrtStatusString(rrt.status))
        self.drawBitmapString(string)

        # Dump an image if we're recording the solution.
        if self.grabPhase == 1:
            self.grabber.grabFrame(verbose=True)

        # Swap the draw buffers to actually display the image.
        glutSwapBuffers()

    # Helper function to draw the robot at a particular state:
    def drawRobotAt(self, pose, colored=True, filled=True):
        self.rrt.robot.xform = pose
        self.rrt.robot.draw(colored=colored, filled=filled)

    # Callback to handle keyboard input.
    def keyboard(self, key, x, y):
        key = key.lower()
        if key == 's':
            self.animating = False
            self.stepRRT()
        if key == '\r' or key == '\n':
            if self.rrt.status <= 0:
                self.resetRRT()
            self.rrt.search()
            glutPostRedisplay()
        if key == 'r':
            self.animating = False
            self.resetRRT()
            glutPostRedisplay()
        if key == 'c':
            self.rrt.useConnectHeuristic = not self.rrt.useConnectHeuristic
            glutPostRedisplay()
        elif key == ' ':
            self.animating = not self.animating
            glutPostRedisplay()
        else:
            GlutApp.keyboard(self, key, x, y)
                
# Very short main function just prints a message and runs the app.
if __name__ == '__main__':
    print "Press 'S' to step, 'R' to reset, 'C' and to toggle connect heuristic."
    print "Enter finishes search, and Space animates."
    a = RRTTransRotApp()
    a.run()


